# ActivitySim
# See full license in LICENSE.txt.
import os
import shutil
import subprocess
import sys

import pytest

if sys.version_info < (3, 7):
    pytest.skip("capture_output introduced in Python 3.7", allow_module_level=True)


def test_help():

    # cp = completed process
    cp = subprocess.run(["activitysim", "-h"], capture_output=True)

    assert "usage: activitysim [-h] [--version]" in str(cp.stdout)


def test_create_help():

    cp = subprocess.run(["activitysim", "create", "-h"], capture_output=True)

    assert "usage: activitysim create [-h] (-l | -e PATH) [-d PATH]" in str(cp.stdout)


def test_create_list():

    cp = subprocess.run(["activitysim", "create", "--list"], capture_output=True)

    assert "Available examples" in str(cp.stdout)
    assert "name: prototype_mtc" in str(cp.stdout)


def test_create_copy():

    target = os.path.join(os.path.dirname(__file__), "test_example")
    cp = subprocess.run(
        [
            "activitysim",
            "create",
            "--example",
            "prototype_mtc",
            "--destination",
            target,
        ],
        capture_output=True,
    )

    assert "copying data ..." in str(cp.stdout)
    assert "copying configs ..." in str(cp.stdout)
    assert "copying configs_mp ..." in str(cp.stdout)
    assert "copying output ..." in str(cp.stdout)

    # replace slashes on windows
    assert str(target).replace("\\\\", "\\") in str(cp.stdout).replace("\\\\", "\\")

    assert os.path.exists(target)
    for folder in ["configs", "configs_mp", "data", "output"]:
        assert os.path.isdir(os.path.join(target, folder))

    # clean up
    shutil.rmtree(target)
    assert not os.path.exists(target)


def test_run():

    cp = subprocess.run(["activitysim", "run"], capture_output=True)

    msg = (
        "Error: please specify either a --working_dir "
        "containing 'configs', 'data', and 'output' "
        "folders or all three of --config, --data, and --output"
    )

    # expect error
    assert msg in str(cp.stderr)


if __name__ == "__main__":

    test_help()
    test_create_help()
    test_create_list()
    test_create_copy()
    test_run()



# ActivitySim
# See full license in LICENSE.txt.
import logging
import os
import warnings

import pandas as pd

from activitysim.abm.models.util import tour_frequency as tf
from activitysim.core import config, expressions, inject, pipeline, tracing
from activitysim.core.input import read_input_table

logger = logging.getLogger(__name__)

SURVEY_TOUR_ID = "external_tour_id"
SURVEY_PARENT_TOUR_ID = "external_parent_tour_id"
SURVEY_PARTICIPANT_ID = "external_participant_id"
ASIM_TOUR_ID = "tour_id"
ASIM_PARENT_TOUR_ID = "parent_tour_id"
REQUIRED_TOUR_COLUMNS = set(["person_id", "tour_category", "tour_type"])


def patch_tour_ids(tours):
    def set_tour_index(tours, parent_tour_num_col, is_joint):
        group_cols = ["person_id", "tour_category", "tour_type"]

        if "parent_tour_num" in tours:
            group_cols += ["parent_tour_num"]

        tours["tour_type_num"] = (
            tours.sort_values(by=group_cols).groupby(group_cols).cumcount() + 1
        )

        return tf.set_tour_index(
            tours, parent_tour_num_col=parent_tour_num_col, is_joint=is_joint
        )

    assert REQUIRED_TOUR_COLUMNS.issubset(
        set(tours.columns)
    ), f"Required columns missing from tours table: {REQUIRED_TOUR_COLUMNS.difference(set(tours.columns))}"

    # replace tour index with asim standard tour_ids (which are based on person_id and tour_type)
    if tours.index.name is not None:
        tours.insert(loc=0, column="legacy_index", value=tours.index)

    # FIXME - for now, only grok simple tours
    assert set(tours.tour_category.unique()).issubset({"mandatory", "non_mandatory"})

    # mandatory tours
    mandatory_tours = set_tour_index(
        tours[tours.tour_category == "mandatory"],
        parent_tour_num_col=None,
        is_joint=False,
    )

    assert mandatory_tours.index.name == "tour_id"

    # FIXME joint tours not implemented
    assert not (tours.tour_category == "joint").any()

    # non_mandatory tours
    non_mandatory_tours = set_tour_index(
        tours[tours.tour_category == "non_mandatory"],
        parent_tour_num_col=None,
        is_joint=False,
    )

    # FIXME atwork tours ot implemented
    assert not (tours.tour_category == "atwork").any()

    patched_tours = pd.concat([mandatory_tours, non_mandatory_tours])
    del patched_tours["tour_type_num"]

    return patched_tours


@inject.step()
def initialize_tours(network_los, households, persons, trace_hh_id):

    trace_label = "initialize_tours"

    tours = read_input_table("tours")

    # FIXME can't use households_sliced injectable as flag like persons table does in case of resume_after.
    # FIXME could just always slice...
    slice_happened = (
        inject.get_injectable("households_sample_size", 0) > 0
        or inject.get_injectable("households_sample_size", 0) > 0
    )
    if slice_happened:
        logger.info("slicing tours %s" % (tours.shape,))
        # keep all persons in the sampled households
        tours = tours[tours.person_id.isin(persons.index)]

    # annotate before patching tour_id to allow addition of REQUIRED_TOUR_COLUMNS defined above
    model_settings = config.read_model_settings("initialize_tours.yaml", mandatory=True)
    expressions.assign_columns(
        df=tours,
        model_settings=model_settings.get("annotate_tours"),
        trace_label=tracing.extend_trace_label(trace_label, "annotate_tours"),
    )

    skip_patch_tour_ids = model_settings.get("skip_patch_tour_ids", False)
    if skip_patch_tour_ids:
        pass
    else:
        tours = patch_tour_ids(tours)
    assert tours.index.name == "tour_id"

    # replace table function with dataframe
    inject.add_table("tours", tours)

    pipeline.get_rn_generator().add_channel("tours", tours)

    tracing.register_traceable_table("tours", tours)

    logger.debug(f"{len(tours.household_id.unique())} unique household_ids in tours")
    logger.debug(f"{len(households.index.unique())} unique household_ids in households")
    assert not tours.index.duplicated().any()

    tours_without_persons = ~tours.person_id.isin(persons.index)
    if tours_without_persons.any():
        logger.error(
            f"{tours_without_persons.sum()} tours out of {len(persons)} without persons\n"
            f"{pd.Series({'person_id': tours_without_persons.index.values})}"
        )
        raise RuntimeError(f"{tours_without_persons.sum()} tours with bad person_id")

    if trace_hh_id:
        tracing.trace_df(tours, label="initialize_tours", warn_if_empty=True)



# ActivitySim
# See full license in LICENSE.txt.
import argparse
import glob
import logging
import os
import struct
import time
import warnings

import yaml

from activitysim.core import inject

logger = logging.getLogger(__name__)

"""
    default injectables
"""


@inject.injectable(cache=True)
def locutor():
    # when multiprocessing, sometimes you only want one process to write trace files
    # mp_tasks overrides this definition to designate a single sub-process as locutor
    return True


@inject.injectable(cache=True)
def configs_dir():
    if not os.path.exists("configs"):
        raise RuntimeError("'configs' directory does not exist")
    return "configs"


@inject.injectable(cache=True)
def data_dir():
    if not os.path.exists("data"):
        raise RuntimeError("'data' directory does not exist")
    return "data"


@inject.injectable(cache=True)
def output_dir():
    if not os.path.exists("output"):
        print(
            f"'output' directory does not exist - current working directory: {os.getcwd()}"
        )
        raise RuntimeError("'output' directory does not exist")
    return "output"


@inject.injectable()
def output_file_prefix():
    return ""


@inject.injectable(cache=True)
def pipeline_file_name(settings):

    pipeline_file_name = settings.get("pipeline_file_name", "pipeline.h5")

    return pipeline_file_name


@inject.injectable()
def rng_base_seed():
    return 0


@inject.injectable(cache=True)
def settings_file_name():
    return "settings.yaml"


@inject.injectable(cache=True)
def settings(settings_file_name):
    settings_dict = read_settings_file(settings_file_name, mandatory=True)

    return settings_dict


# def testing():
#
#     assert ("pytest" in sys.modules) == ("PYTEST_CURRENT_TEST" in os.environ)
#     return "PYTEST_CURRENT_TEST" in os.environ


def get_cache_dir():
    """
    return path of cache directory in output_dir (creating it, if need be)

    cache directory is used to store
        skim memmaps created by skim+dict_factories
        tvpb tap_tap table cache

    Returns
    -------
    str path
    """
    cache_dir = setting("cache_dir", default=None)
    if cache_dir is None:
        cache_dir = setting(
            "cache_dir", os.path.join(inject.get_injectable("output_dir"), "cache")
        )

    if not os.path.isdir(cache_dir):
        os.mkdir(cache_dir)
    assert os.path.isdir(cache_dir)

    return cache_dir


def setting(key, default=None):
    return inject.get_injectable("settings").get(key, default)


def override_setting(key, value):
    new_settings = inject.get_injectable("settings")
    new_settings[key] = value
    inject.add_injectable("settings", new_settings)


def get_global_constants():
    """
    Read global constants from settings file

    Returns
    -------
    constants : dict
        dictionary of constants to add to locals for use by expressions in model spec
    """
    return read_settings_file("constants.yaml", mandatory=False)


def read_model_settings(file_name, mandatory=False):
    """

    Parameters
    ----------
    file_name : str
        yaml file name
    mandatory : bool
        throw error if file empty or not found
    Returns
    -------

    """

    model_settings = read_settings_file(file_name, mandatory=mandatory)

    return model_settings


def future_model_settings(model_name, model_settings, future_settings):
    """
    Warn users of new required model settings, and substitute default values

    Parameters
    ----------
    model_name: str
        name of model
    model_settings: dict
        model_settings from settigns file
    future_settings: dict
        default values for new required settings

    Returns
    -------
        dict
            model_settings with any missing future_settings added

    """
    model_settings = model_settings.copy()
    for key, setting in future_settings.items():
        if key not in model_settings.keys():
            warnings.warn(
                f"Setting '{key}' not found in {model_name} model settings."
                f"Replacing with default value: {setting}."
                f"This setting will be required in future versions",
                FutureWarning,
            )
            model_settings[key] = setting

    return model_settings


def get_model_constants(model_settings):
    """
    Read constants from model settings file

    Returns
    -------
    constants : dict
        dictionary of constants to add to locals for use by expressions in model spec
    """
    return model_settings.get("CONSTANTS", {})


def get_logit_model_settings(model_settings):
    """
    Read nest spec (for nested logit) from model settings file

    Returns
    -------
    nests : dict
        dictionary specifying nesting structure and nesting coefficients

    constants : dict
        dictionary of constants to add to locals for use by expressions in model spec
    """
    nests = None

    if model_settings is not None:

        # default to MNL
        logit_type = model_settings.get("LOGIT_TYPE", "MNL")

        if logit_type not in ["NL", "MNL"]:
            logger.error("Unrecognized logit type '%s'" % logit_type)
            raise RuntimeError("Unrecognized logit type '%s'" % logit_type)

        if logit_type == "NL":
            nests = model_settings.get("NESTS", None)
            if nests is None:
                logger.error("No NEST found in model spec for NL model type")
                raise RuntimeError("No NEST found in model spec for NL model type")

    return nests


def build_output_file_path(file_name, use_prefix=None):
    output_dir = inject.get_injectable("output_dir")

    if use_prefix:
        file_name = "%s-%s" % (use_prefix, file_name)

    file_path = os.path.join(output_dir, file_name)

    return file_path


def cascading_input_file_path(
    file_name, dir_list_injectable_name, mandatory=True, allow_glob=False
):

    dir_paths = inject.get_injectable(dir_list_injectable_name)
    dir_paths = [dir_paths] if isinstance(dir_paths, str) else dir_paths

    file_path = None
    for dir in dir_paths:
        p = os.path.join(dir, file_name)
        if os.path.isfile(p):
            file_path = p
            break

        if allow_glob and len(glob.glob(p)) > 0:
            file_path = p
            break

    if mandatory and not file_path:
        raise RuntimeError(
            "file_path %s: file '%s' not in %s"
            % (dir_list_injectable_name, file_name, dir_paths)
        )

    return file_path


def data_file_path(file_name, mandatory=True, allow_glob=False):

    return cascading_input_file_path(
        file_name, "data_dir", mandatory=mandatory, allow_glob=allow_glob
    )


def expand_input_file_list(input_files):
    """
    expand list by unglobbing globs globs
    """

    # be nice and accept a string as well as a list of strings
    if isinstance(input_files, str):
        input_files = [input_files]

    expanded_files = []
    ungroked_files = 0

    for file_name in input_files:

        file_name = data_file_path(file_name, allow_glob=True)

        if os.path.isfile(file_name):
            expanded_files.append(file_name)
            continue

        if os.path.isdir(file_name):
            logger.warning(
                "WARNING: expand_input_file_list skipping directory: "
                "(use glob instead): %s",
                file_name,
            )
            ungroked_files += 1
            continue

        # - glob
        logger.debug(f"expand_input_file_list trying {file_name} as glob")
        globbed_files = glob.glob(file_name)
        for globbed_file in globbed_files:
            if os.path.isfile(globbed_file):
                expanded_files.append(globbed_file)
            else:
                logger.warning(
                    "WARNING: expand_input_file_list skipping: " "(does not grok) %s",
                    file_name,
                )
                ungroked_files += 1

        if len(globbed_files) == 0:
            logger.warning(
                "WARNING: expand_input_file_list file/glob not found: %s", file_name
            )

    assert ungroked_files == 0, f"{ungroked_files} ungroked file names"

    return sorted(expanded_files)


def config_file_path(file_name, mandatory=True):

    return cascading_input_file_path(file_name, "configs_dir", mandatory)


def output_file_path(file_name):

    prefix = inject.get_injectable("output_file_prefix", None)
    return build_output_file_path(file_name, use_prefix=prefix)


def trace_file_path(file_name):

    output_dir = inject.get_injectable("output_dir")

    # - check for trace subfolder, create it if missing
    trace_dir = os.path.join(output_dir, "trace")
    if not os.path.exists(trace_dir):
        os.makedirs(trace_dir)

    # construct a unique tail string from the time
    # this is a convenience for opening multiple similarly named trace files
    tail = hex(struct.unpack("<Q", struct.pack("<d", time.time()))[0])[-6:]

    file_parts = file_name.split(".")

    file_path = os.path.join(trace_dir, *file_parts[:-1]) + f"-{tail}.{file_parts[-1]}"
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    return file_path


def log_file_path(file_name, prefix=True):

    output_dir = inject.get_injectable("output_dir")

    # - check if running asv and if so, log to commit-specific subfolder
    asv_commit = os.environ.get("ASV_COMMIT", None)
    if asv_commit:
        output_dir = os.path.join(output_dir, f"log-{asv_commit}")
        os.makedirs(output_dir, exist_ok=True)

    # - check for optional log subfolder
    if os.path.exists(os.path.join(output_dir, "log")):
        output_dir = os.path.join(output_dir, "log")

    # - check for optional process name prefix
    prefix = prefix and inject.get_injectable("log_file_prefix", None)
    if prefix:
        file_name = "%s-%s" % (prefix, file_name)

    file_path = os.path.join(output_dir, file_name)

    return file_path


def open_log_file(file_name, mode, header=None, prefix=False):

    file_path = log_file_path(file_name, prefix)

    want_header = header and not os.path.exists(file_path)

    f = open(file_path, mode)

    if want_header:
        assert mode in [
            "a",
            "w",
        ], f"open_log_file: header requested but mode was {mode}"
        print(header, file=f)

    return f


def pipeline_file_path(file_name):

    prefix = inject.get_injectable("pipeline_file_prefix", None)
    return build_output_file_path(file_name, use_prefix=prefix)


class SettingsFileNotFound(Exception):
    def __init__(self, file_name, configs_dir):
        self.file_name = file_name
        self.configs_dir = configs_dir

    def __str__(self):
        return repr(f"Settings file '{self.file_name}' not found in {self.configs_dir}")


def read_settings_file(
    file_name, mandatory=True, include_stack=False, configs_dir_list=None
):
    """

    look for first occurence of yaml file named <file_name> in directories in configs_dir list,
    read settings from yaml file and return as dict.

    Settings file may contain directives that affect which file settings are returned:

    inherit_settings: boolean
        backfill settings in the current file with values from the next settings file in configs_dir list
    include_settings: string <include_file_name>
        read settings from specified include_file in place of the current file settings
        (to avoid confusion, this directive must appea ALONE in fiel, without any additional settings or directives.)

    Parameters
    ----------
    file_name
    mandatory: booelan
        if true, raise SettingsFileNotFound exception if no settings file, otherwise return empty dict
    include_stack: boolean or list
        only used for recursive calls to provide list of files included so far to detect cycles

    Returns: dict
        settings from speciified settings file/s
    -------

    """

    def backfill_settings(settings, backfill):
        new_settings = backfill.copy()
        new_settings.update(settings)
        return new_settings

    if configs_dir_list is None:
        configs_dir_list = inject.get_injectable("configs_dir")
        configs_dir_list = (
            [configs_dir_list]
            if isinstance(configs_dir_list, str)
            else configs_dir_list
        )
        assert isinstance(configs_dir_list, list)
        assert len(configs_dir_list) == len(
            set(configs_dir_list)
        ), f"repeating file names not allowed in config_dir list: {configs_dir_list}"

    if not file_name.lower().endswith(".yaml"):
        file_name = "%s.yaml" % (file_name,)

    inheriting = False
    settings = {}
    if isinstance(include_stack, list):
        source_file_paths = include_stack.copy()
    else:
        source_file_paths = []
    for dir in configs_dir_list:
        file_path = os.path.join(dir, file_name)
        if os.path.exists(file_path):
            if inheriting:
                # we must be inheriting
                logger.debug(
                    "inheriting additional settings for %s from %s"
                    % (file_name, file_path)
                )
                inheriting = True

            assert (
                file_path not in source_file_paths
            ), f"read_settings_file - recursion in reading 'file_path' after loading: {source_file_paths}"

            with open(file_path) as f:

                s = yaml.load(f, Loader=yaml.SafeLoader)
                if s is None:
                    s = {}

            settings = backfill_settings(settings, s)

            # maintain a list of files we read from to improve error message when an expected setting is not found
            source_file_paths += [file_path]

            include_file_name = s.get("include_settings", False)
            if include_file_name:
                # FIXME - prevent users from creating borgesian garden of branching paths?
                # There is a lot of opportunity for confusion if this feature were over-used
                # Maybe we insist that a file with an include directive is the 'end of the road'
                # essentially the current settings firle is an alias for the included file
                if len(s) > 1:
                    logger.error(
                        f"'include_settings' must appear alone in settings file."
                    )
                    additional_settings = list(
                        set(s.keys()).difference({"include_settings"})
                    )
                    logger.error(
                        f"Unexpected additional settings: {additional_settings}"
                    )
                    raise RuntimeError(
                        f"'include_settings' must appear alone in settings file."
                    )

                logger.debug(
                    "including settings for %s from %s" % (file_name, include_file_name)
                )

                # recursive call to read included file INSTEAD of the file  with include_settings sepcified
                s, source_file_paths = read_settings_file(
                    include_file_name, mandatory=True, include_stack=source_file_paths
                )

                # FIXME backfill with the included file
                settings = backfill_settings(settings, s)

            # we are done as soon as we read one file successfully
            # unless if inherit_settings is set to true in this file

            if not s.get("inherit_settings", False):
                break

            # if inheriting, continue and backfill settings from the next existing settings file configs_dir_list

            inherit_settings = s.get("inherit_settings")
            if isinstance(inherit_settings, str):
                inherit_file_name = inherit_settings
                assert (
                    os.path.join(dir, inherit_file_name) not in source_file_paths
                ), f"circular inheritance of {inherit_file_name}: {source_file_paths}: "
                # make a recursive call to switch inheritance chain to specified file

                logger.debug(
                    "inheriting additional settings for %s from %s"
                    % (file_name, inherit_file_name)
                )
                s, source_file_paths = read_settings_file(
                    inherit_file_name,
                    mandatory=True,
                    include_stack=source_file_paths,
                    configs_dir_list=configs_dir_list,
                )

                # backfill with the inherited file
                settings = backfill_settings(settings, s)
                break  # break the current inheritance chain (not as bad luck as breaking a chain-letter chain?...)

    if len(source_file_paths) > 0:
        settings["source_file_paths"] = source_file_paths

    if mandatory and not settings:
        raise SettingsFileNotFound(file_name, configs_dir_list)

    if include_stack:
        # if we were called recursively, return an updated list of source_file_paths
        return settings, source_file_paths

    else:
        return settings


def base_settings_file_path(file_name):
    """

    Parameters
    ----------
    file_name

    Returns
    -------
        path to base settings file or None if not found
    """

    if not file_name.lower().endswith(".yaml"):
        file_name = "%s.yaml" % (file_name,)

    configs_dir = inject.get_injectable("configs_dir")
    configs_dir = [configs_dir] if isinstance(configs_dir, str) else configs_dir

    for dir in configs_dir:
        file_path = os.path.join(dir, file_name)
        if os.path.exists(file_path):
            return file_path

    raise RuntimeError("base_settings_file %s not found" % file_name)


def filter_warnings():
    """
    set warning filter to 'strict' if specified in settings
    """

    if setting("strict", False):  # noqa: E402
        warnings.filterwarnings("error", category=Warning)
        warnings.filterwarnings(
            "default", category=PendingDeprecationWarning, module="future"
        )
        warnings.filterwarnings("default", category=FutureWarning, module="pandas")
        warnings.filterwarnings("default", category=RuntimeWarning, module="numpy")

    # pandas pytables.py __getitem__ (e.g. df = store['any_string'])
    # indirectly raises tables DeprecationWarning: tostring() is deprecated. Use tobytes() instead.
    warnings.filterwarnings(
        "ignore", category=DeprecationWarning, module="tables", message="tostring"
    )

    #   File "tables/hdf5extension.pyx", line 1450, in tables.hdf5extension.Array._open_array
    # DeprecationWarning: `np.object` is a deprecated alias for the builtin `object`.
    # Deprecated in NumPy 1.20;
    warnings.filterwarnings(
        "ignore",
        category=DeprecationWarning,
        module="tables",
        message="`np.object` is a deprecated alias",
    )

    # beginning pandas version 1.3, various places emit a PerformanceWarning that is
    # caught in the "strict" filter above, but which are currently unavoidable for complex models.
    # These warning are left as warnings as an invitation for future enhancement.
    from pandas.errors import PerformanceWarning

    warnings.filterwarnings("default", category=PerformanceWarning)


def handle_standard_args(parser=None):

    from activitysim.cli import run

    warnings.warn(
        "config.handle_standard_args() has been moved to the command line "
        "module and will be removed in future versions.",
        FutureWarning,
    )

    if parser is None:
        parser = argparse.ArgumentParser()

    run.add_run_args(parser)
    args = parser.parse_args()
    run.handle_standard_args(args)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.abm.tables.size_terms import tour_destination_size_terms
from activitysim.core import config, inject, los, pipeline, simulate, tracing
from activitysim.core.interaction_sample import interaction_sample
from activitysim.core.interaction_sample_simulate import interaction_sample_simulate
from activitysim.core.util import reindex

from . import logsums as logsum

logger = logging.getLogger(__name__)
DUMP = False


class SizeTermCalculator(object):
    """
    convenience object to provide size_terms for a selector (e.g. non_mandatory)
    for various segments (e.g. tour_type or purpose)
    returns size terms for specified segment in df or series form
    """

    def __init__(self, size_term_selector):

        # do this once so they can request size_terms for various segments (tour_type or purpose)
        land_use = inject.get_table("land_use")
        size_terms = inject.get_injectable("size_terms")
        self.destination_size_terms = tour_destination_size_terms(
            land_use, size_terms, size_term_selector
        )

        assert not self.destination_size_terms.isna().any(axis=None)

    # def omnibus_size_terms_df(self):
    #     return self.destination_size_terms

    def dest_size_terms_df(self, segment_name, trace_label):
        # return size terms as df with one column named 'size_term'
        # convenient if creating or merging with alts

        size_terms = self.destination_size_terms[[segment_name]].copy()
        size_terms.columns = ["size_term"]

        # FIXME - no point in considering impossible alternatives (where dest size term is zero)
        logger.debug(
            f"SizeTermCalculator dropping {(~(size_terms.size_term > 0)).sum()} "
            f"of {len(size_terms)} rows where size_term is zero for {segment_name}"
        )
        size_terms = size_terms[size_terms.size_term > 0]

        if len(size_terms) == 0:
            logger.warning(
                f"SizeTermCalculator: no zones with non-zero size terms for {segment_name} in {trace_label}"
            )

        return size_terms

    # def dest_size_terms_series(self, segment_name):
    #     # return size terms as as series
    #     # convenient (and no copy overhead) if reindexing and assigning into alts column
    #     return self.destination_size_terms[segment_name]


def _destination_sample(
    spec_segment_name,
    choosers,
    destination_size_terms,
    skims,
    estimator,
    model_settings,
    alt_dest_col_name,
    chunk_size,
    chunk_tag,
    trace_label,
):

    model_spec = simulate.spec_for_segment(
        model_settings,
        spec_id="SAMPLE_SPEC",
        segment_name=spec_segment_name,
        estimator=estimator,
    )

    logger.info("running %s with %d tours", trace_label, len(choosers))

    sample_size = model_settings["SAMPLE_SIZE"]
    if config.setting("disable_destination_sampling", False) or (
        estimator and estimator.want_unsampled_alternatives
    ):
        # FIXME interaction_sample will return unsampled complete alternatives with probs and pick_count
        logger.info(
            "Estimation mode for %s using unsampled alternatives short_circuit_choices"
            % (trace_label,)
        )
        sample_size = 0

    locals_d = {"skims": skims}
    constants = config.get_model_constants(model_settings)
    if constants is not None:
        locals_d.update(constants)

    log_alt_losers = config.setting("log_alt_losers", False)

    choices = interaction_sample(
        choosers,
        alternatives=destination_size_terms,
        sample_size=sample_size,
        alt_col_name=alt_dest_col_name,
        log_alt_losers=log_alt_losers,
        spec=model_spec,
        skims=skims,
        locals_d=locals_d,
        chunk_size=chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
    )

    # remember person_id in chosen alts so we can merge with persons in subsequent steps
    # (broadcasts person_id onto all alternatives sharing the same tour_id index value)
    choices["person_id"] = choosers.person_id

    return choices


def destination_sample(
    spec_segment_name,
    choosers,
    model_settings,
    network_los,
    destination_size_terms,
    estimator,
    chunk_size,
    trace_label,
):

    chunk_tag = "tour_destination.sample"

    # create wrapper with keys for this lookup
    # the skims will be available under the name "skims" for any @ expressions
    skim_origin_col_name = model_settings["CHOOSER_ORIG_COL_NAME"]
    skim_dest_col_name = destination_size_terms.index.name
    # (logit.interaction_dataset suffixes duplicate chooser column with '_chooser')
    if skim_origin_col_name == skim_dest_col_name:
        skim_origin_col_name = f"{skim_origin_col_name}_chooser"

    skim_dict = network_los.get_default_skim_dict()
    skims = skim_dict.wrap(skim_origin_col_name, skim_dest_col_name)

    # the name of the dest column to be returned in choices
    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]

    choices = _destination_sample(
        spec_segment_name,
        choosers,
        destination_size_terms,
        skims,
        estimator,
        model_settings,
        alt_dest_col_name,
        chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
    )

    return choices


# temp column names for presampling
DEST_MAZ = "dest_MAZ"
DEST_TAZ = "dest_TAZ"
ORIG_TAZ = "TAZ"  # likewise a temp, but if already in choosers, we assume we can use it opportunistically


def map_maz_to_taz(s, network_los):
    maz_to_taz = network_los.maz_taz_df[["MAZ", "TAZ"]].set_index("MAZ").TAZ
    return s.map(maz_to_taz)


def aggregate_size_terms(dest_size_terms, network_los):
    #
    # aggregate MAZ_size_terms to TAZ_size_terms
    #

    MAZ_size_terms = dest_size_terms.copy()

    # add crosswalk DEST_TAZ column to MAZ_size_terms
    MAZ_size_terms[DEST_TAZ] = map_maz_to_taz(MAZ_size_terms.index, network_los)

    # aggregate to TAZ
    TAZ_size_terms = MAZ_size_terms.groupby(DEST_TAZ).agg({"size_term": "sum"})
    TAZ_size_terms[DEST_TAZ] = TAZ_size_terms.index
    assert not TAZ_size_terms["size_term"].isna().any()

    #           size_term
    # dest_TAZ
    # 2              45.0
    # 3              44.0
    # 4              59.0

    # add crosswalk DEST_TAZ column to MAZ_size_terms
    # MAZ_size_terms = MAZ_size_terms.sort_values([DEST_TAZ, 'size_term'])  # maybe helpful for debugging
    MAZ_size_terms = MAZ_size_terms[[DEST_TAZ, "size_term"]].reset_index(drop=False)
    MAZ_size_terms = MAZ_size_terms.sort_values([DEST_TAZ, "zone_id"]).reset_index(
        drop=True
    )

    #       zone_id  dest_TAZ  size_term
    # 0        6097         2       10.0
    # 1       16421         2       13.0
    # 2       24251         3       14.0

    # print(f"TAZ_size_terms ({TAZ_size_terms.shape})\n{TAZ_size_terms}")
    # print(f"MAZ_size_terms ({MAZ_size_terms.shape})\n{MAZ_size_terms}")

    return MAZ_size_terms, TAZ_size_terms


def choose_MAZ_for_TAZ(taz_sample, MAZ_size_terms, trace_label):
    """
    Convert taz_sample table with TAZ zone sample choices to a table with a MAZ zone chosen for each TAZ
    choose MAZ probabilistically (proportionally by size_term) from set of MAZ zones in parent TAZ

    Parameters
    ----------
    taz_sample: dataframe with duplicated index <chooser_id_col> and columns: <DEST_TAZ>, prob, pick_count
    MAZ_size_terms: dataframe with duplicated index <chooser_id_col> and columns: zone_id, dest_TAZ, size_term

    Returns
    -------
    dataframe with with duplicated index <chooser_id_col> and columns: <DEST_MAZ>, prob, pick_count
    """

    # print(f"taz_sample\n{taz_sample}")
    #           dest_TAZ      prob  pick_count  person_id
    # tour_id
    # 542963          18  0.004778           1      13243
    # 542963          53  0.004224           2      13243
    # 542963          59  0.008628           1      13243

    trace_hh_id = inject.get_injectable("trace_hh_id", None)
    have_trace_targets = trace_hh_id and tracing.has_trace_targets(taz_sample)
    if have_trace_targets:
        trace_label = tracing.extend_trace_label(trace_label, "choose_MAZ_for_TAZ")

        CHOOSER_ID = (
            taz_sample.index.name
        )  # zone_id for tours, but person_id for location choice
        assert CHOOSER_ID is not None

        # write taz choices, pick_counts, probs
        trace_targets = tracing.trace_targets(taz_sample)
        tracing.trace_df(
            taz_sample[trace_targets],
            label=tracing.extend_trace_label(trace_label, "taz_sample"),
            transpose=False,
        )

    # redupe taz_sample[[DEST_TAZ, 'prob']] using pick_count to repeat rows
    taz_choices = taz_sample[[DEST_TAZ, "prob"]].reset_index(drop=False)
    taz_choices = taz_choices.reindex(
        taz_choices.index.repeat(taz_sample.pick_count)
    ).reset_index(drop=True)
    taz_choices = taz_choices.rename(columns={"prob": "TAZ_prob"})

    # print(f"taz_choices\n{taz_choices}")
    #        tour_id  dest_TAZ  TAZ_prob
    # 0       542963        18  0.004778
    # 1       542963        53  0.004224
    # 2       542963        53  0.004224
    # 3       542963        59  0.008628

    # print(f"MAZ_size_terms\n{MAZ_size_terms}")
    #       zone_id  dest_TAZ  size_term
    # 0        6097         2      7.420
    # 1       16421         2      9.646
    # 2       24251         2     10.904

    # just to make it clear we are siloing choices by chooser_id
    chooser_id_col = (
        taz_sample.index.name
    )  # should be canonical chooser index name (e.g. 'person_id')

    # for random_for_df, we need df with de-duplicated chooser canonical index
    chooser_df = pd.DataFrame(index=taz_sample.index[~taz_sample.index.duplicated()])
    num_choosers = len(chooser_df)
    assert chooser_df.index.name == chooser_id_col

    # to make choices, <taz_sample_size> rands for each chooser (one rand for each sampled TAZ)
    # taz_sample_size will be model_settings['SAMPLE_SIZE'] samples, except if we are estimating
    taz_sample_size = taz_choices.groupby(chooser_id_col)[DEST_TAZ].count().max()

    # taz_choices index values should be contiguous
    assert (
        taz_choices[chooser_id_col] == np.repeat(chooser_df.index, taz_sample_size)
    ).all()

    # we need to choose a MAZ for each DEST_TAZ choice
    # probability of choosing MAZ based on MAZ size_term fraction of TAZ total
    # there will be a different set (and number) of candidate MAZs for each TAZ
    # (preserve index, which will have duplicates as result of join)
    # maz_sizes.index is the integer offset into taz_choices of the taz for which the maz_size row is a candidate)
    maz_sizes = pd.merge(
        taz_choices[[chooser_id_col, DEST_TAZ]].reset_index(),
        MAZ_size_terms,
        how="left",
        on=DEST_TAZ,
    ).set_index("index")

    #         tour_id  dest_TAZ  zone_id  size_term
    # index
    # 0        542963        18      498     12.130
    # 0        542963        18     7696     18.550
    # 0        542963        18    15431      8.678
    # 0        542963        18    21429     29.938
    # 1        542963        53    17563     34.252

    if have_trace_targets:
        # write maz_sizes: maz_sizes[index,tour_id,dest_TAZ,zone_id,size_term]

        maz_sizes_trace_targets = tracing.trace_targets(maz_sizes, slicer=CHOOSER_ID)
        trace_maz_sizes = maz_sizes[maz_sizes_trace_targets]
        tracing.trace_df(
            trace_maz_sizes,
            label=tracing.extend_trace_label(trace_label, "maz_sizes"),
            transpose=False,
        )

    # number of DEST_TAZ candidates per chooser
    maz_counts = maz_sizes.groupby(maz_sizes.index).size().values

    # max number of MAZs for any TAZ
    max_maz_count = maz_counts.max()

    # offsets of the first and last rows of each chooser in sparse interaction_utilities
    last_row_offsets = maz_counts.cumsum()
    first_row_offsets = np.insert(last_row_offsets[:-1], 0, 0)

    # repeat the row offsets once for each dummy utility to insert
    # (we want to insert dummy utilities at the END of the list of alternative utilities)
    # inserts is a list of the indices at which we want to do the insertions
    inserts = np.repeat(last_row_offsets, max_maz_count - maz_counts)

    # insert zero filler to pad each alternative set to same size
    padded_maz_sizes = np.insert(maz_sizes.size_term.values, inserts, 0.0).reshape(
        -1, max_maz_count
    )

    # prob array with one row TAZ_choice, one column per alternative
    row_sums = padded_maz_sizes.sum(axis=1)
    maz_probs = np.divide(padded_maz_sizes, row_sums.reshape(-1, 1))
    assert maz_probs.shape == (num_choosers * taz_sample_size, max_maz_count)

    rands = pipeline.get_rn_generator().random_for_df(chooser_df, n=taz_sample_size)
    rands = rands.reshape(-1, 1)
    assert len(rands) == num_choosers * taz_sample_size
    assert len(rands) == maz_probs.shape[0]

    # make choices
    # positions is array with the chosen alternative represented as a column index in probs
    # which is an integer between zero and max_maz_count
    positions = np.argmax((maz_probs.cumsum(axis=1) - rands) > 0.0, axis=1)

    # shouldn't have chosen any of the dummy pad positions
    assert (positions < maz_counts).all()

    taz_choices[DEST_MAZ] = maz_sizes["zone_id"].take(positions + first_row_offsets)
    taz_choices["MAZ_prob"] = maz_probs[np.arange(maz_probs.shape[0]), positions]
    taz_choices["prob"] = taz_choices["TAZ_prob"] * taz_choices["MAZ_prob"]

    if have_trace_targets:

        taz_choices_trace_targets = tracing.trace_targets(
            taz_choices, slicer=CHOOSER_ID
        )
        trace_taz_choices_df = taz_choices[taz_choices_trace_targets]
        tracing.trace_df(
            trace_taz_choices_df,
            label=tracing.extend_trace_label(trace_label, "taz_choices"),
            transpose=False,
        )

        lhs_df = trace_taz_choices_df[[CHOOSER_ID, DEST_TAZ]]
        alt_dest_columns = [f"dest_maz_{c}" for c in range(max_maz_count)]

        # following the same logic as the full code, but for trace cutout
        trace_maz_counts = maz_counts[taz_choices_trace_targets]
        trace_last_row_offsets = maz_counts[taz_choices_trace_targets].cumsum()
        trace_inserts = np.repeat(
            trace_last_row_offsets, max_maz_count - trace_maz_counts
        )

        # trace dest_maz_alts
        padded_maz_sizes = np.insert(
            trace_maz_sizes[CHOOSER_ID].values, trace_inserts, 0.0
        ).reshape(-1, max_maz_count)
        df = pd.DataFrame(
            data=padded_maz_sizes,
            columns=alt_dest_columns,
            index=trace_taz_choices_df.index,
        )
        df = pd.concat([lhs_df, df], axis=1)
        tracing.trace_df(
            df,
            label=tracing.extend_trace_label(trace_label, "dest_maz_alts"),
            transpose=False,
        )

        # trace dest_maz_size_terms
        padded_maz_sizes = np.insert(
            trace_maz_sizes["size_term"].values, trace_inserts, 0.0
        ).reshape(-1, max_maz_count)
        df = pd.DataFrame(
            data=padded_maz_sizes,
            columns=alt_dest_columns,
            index=trace_taz_choices_df.index,
        )
        df = pd.concat([lhs_df, df], axis=1)
        tracing.trace_df(
            df,
            label=tracing.extend_trace_label(trace_label, "dest_maz_size_terms"),
            transpose=False,
        )

        # trace dest_maz_probs
        df = pd.DataFrame(
            data=maz_probs[taz_choices_trace_targets],
            columns=alt_dest_columns,
            index=trace_taz_choices_df.index,
        )
        df = pd.concat([lhs_df, df], axis=1)
        df["rand"] = rands[taz_choices_trace_targets]
        tracing.trace_df(
            df,
            label=tracing.extend_trace_label(trace_label, "dest_maz_probs"),
            transpose=False,
        )

    taz_choices = taz_choices.drop(columns=["TAZ_prob", "MAZ_prob"])
    taz_choices = taz_choices.groupby([chooser_id_col, DEST_MAZ]).agg(
        prob=("prob", "max"), pick_count=("prob", "count")
    )

    taz_choices.reset_index(level=DEST_MAZ, inplace=True)

    return taz_choices


def destination_presample(
    spec_segment_name,
    choosers,
    model_settings,
    network_los,
    destination_size_terms,
    estimator,
    chunk_size,
    trace_label,
):

    trace_label = tracing.extend_trace_label(trace_label, "presample")
    chunk_tag = "tour_destination.presample"

    logger.info(f"{trace_label} location_presample")

    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]
    assert DEST_TAZ != alt_dest_col_name

    MAZ_size_terms, TAZ_size_terms = aggregate_size_terms(
        destination_size_terms, network_los
    )

    orig_maz = model_settings["CHOOSER_ORIG_COL_NAME"]
    assert orig_maz in choosers
    if ORIG_TAZ not in choosers:
        choosers[ORIG_TAZ] = map_maz_to_taz(choosers[orig_maz], network_los)

    # create wrapper with keys for this lookup - in this case there is a HOME_TAZ in the choosers
    # and a DEST_TAZ in the alternatives which get merged during interaction
    # the skims will be available under the name "skims" for any @ expressions
    skim_dict = network_los.get_skim_dict("taz")
    skims = skim_dict.wrap(ORIG_TAZ, DEST_TAZ)

    taz_sample = _destination_sample(
        spec_segment_name,
        choosers,
        TAZ_size_terms,
        skims,
        estimator,
        model_settings,
        DEST_TAZ,
        chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
    )

    # choose a MAZ for each DEST_TAZ choice, choice probability based on MAZ size_term fraction of TAZ total
    maz_choices = choose_MAZ_for_TAZ(taz_sample, MAZ_size_terms, trace_label)

    assert DEST_MAZ in maz_choices
    maz_choices = maz_choices.rename(columns={DEST_MAZ: alt_dest_col_name})

    return maz_choices


def run_destination_sample(
    spec_segment_name,
    tours,
    persons_merged,
    model_settings,
    network_los,
    destination_size_terms,
    estimator,
    chunk_size,
    trace_label,
):

    # FIXME - MEMORY HACK - only include columns actually used in spec (omit them pre-merge)
    chooser_columns = model_settings["SIMULATE_CHOOSER_COLUMNS"]
    persons_merged = persons_merged[
        [c for c in persons_merged.columns if c in chooser_columns]
    ]
    tours = tours[
        [c for c in tours.columns if c in chooser_columns or c == "person_id"]
    ]
    choosers = pd.merge(
        tours, persons_merged, left_on="person_id", right_index=True, how="left"
    )

    # interaction_sample requires that choosers.index.is_monotonic_increasing
    if not choosers.index.is_monotonic_increasing:
        logger.debug(
            f"run_destination_sample {trace_label} sorting choosers because not monotonic_increasing"
        )
        choosers = choosers.sort_index()

    # by default, enable presampling for multizone systems, unless they disable it in settings file
    pre_sample_taz = not (network_los.zone_system == los.ONE_ZONE)
    if pre_sample_taz and not config.setting("want_dest_choice_presampling", True):
        pre_sample_taz = False
        logger.info(
            f"Disabled destination zone presampling for {trace_label} "
            f"because 'want_dest_choice_presampling' setting is False"
        )

    if pre_sample_taz:

        logger.info(
            "Running %s destination_presample with %d tours" % (trace_label, len(tours))
        )

        choices = destination_presample(
            spec_segment_name,
            choosers,
            model_settings,
            network_los,
            destination_size_terms,
            estimator,
            chunk_size,
            trace_label,
        )

    else:
        choices = destination_sample(
            spec_segment_name,
            choosers,
            model_settings,
            network_los,
            destination_size_terms,
            estimator,
            chunk_size,
            trace_label,
        )

    # remember person_id in chosen alts so we can merge with persons in subsequent steps
    # (broadcasts person_id onto all alternatives sharing the same tour_id index value)
    choices["person_id"] = tours.person_id

    return choices


def run_destination_logsums(
    tour_purpose,
    persons_merged,
    destination_sample,
    model_settings,
    network_los,
    chunk_size,
    trace_label,
):
    """
    add logsum column to existing tour_destination_sample table

    logsum is calculated by running the mode_choice model for each sample (person, dest_zone_id) pair
    in destination_sample, and computing the logsum of all the utilities

    +-----------+--------------+----------------+------------+----------------+
    | person_id | dest_zone_id | rand           | pick_count | logsum (added) |
    +===========+==============+================+============+================+
    | 23750     |  14          | 0.565502716034 | 4          |  1.85659498857 |
    +-----------+--------------+----------------+------------+----------------+
    + 23750     | 16           | 0.711135838871 | 6          | 1.92315598631  |
    +-----------+--------------+----------------+------------+----------------+
    + ...       |              |                |            |                |
    +-----------+--------------+----------------+------------+----------------+
    | 23751     | 12           | 0.408038878552 | 1          | 2.40612135416  |
    +-----------+--------------+----------------+------------+----------------+
    | 23751     | 14           | 0.972732479292 | 2          |  1.44009018355 |
    +-----------+--------------+----------------+------------+----------------+
    """

    logsum_settings = config.read_model_settings(model_settings["LOGSUM_SETTINGS"])

    chunk_tag = "tour_destination.logsums"

    # FIXME - MEMORY HACK - only include columns actually used in spec
    persons_merged = logsum.filter_chooser_columns(
        persons_merged, logsum_settings, model_settings
    )

    # merge persons into tours
    choosers = pd.merge(
        destination_sample,
        persons_merged,
        left_on="person_id",
        right_index=True,
        how="left",
    )

    logger.info("Running %s with %s rows", trace_label, len(choosers))

    tracing.dump_df(DUMP, persons_merged, trace_label, "persons_merged")
    tracing.dump_df(DUMP, choosers, trace_label, "choosers")

    logsums = logsum.compute_logsums(
        choosers,
        tour_purpose,
        logsum_settings,
        model_settings,
        network_los,
        chunk_size,
        chunk_tag,
        trace_label,
    )

    destination_sample["mode_choice_logsum"] = logsums

    return destination_sample


def run_destination_simulate(
    spec_segment_name,
    tours,
    persons_merged,
    destination_sample,
    want_logsums,
    model_settings,
    network_los,
    destination_size_terms,
    estimator,
    chunk_size,
    trace_label,
):
    """
    run destination_simulate on tour_destination_sample
    annotated with mode_choice logsum to select a destination from sample alternatives
    """
    chunk_tag = "tour_destination.simulate"

    model_spec = simulate.spec_for_segment(
        model_settings,
        spec_id="SPEC",
        segment_name=spec_segment_name,
        estimator=estimator,
    )

    # FIXME - MEMORY HACK - only include columns actually used in spec (omit them pre-merge)
    chooser_columns = model_settings["SIMULATE_CHOOSER_COLUMNS"]
    persons_merged = persons_merged[
        [c for c in persons_merged.columns if c in chooser_columns]
    ]
    tours = tours[
        [c for c in tours.columns if c in chooser_columns or c == "person_id"]
    ]
    choosers = pd.merge(
        tours, persons_merged, left_on="person_id", right_index=True, how="left"
    )

    # interaction_sample requires that choosers.index.is_monotonic_increasing
    if not choosers.index.is_monotonic_increasing:
        logger.debug(
            f"run_destination_simulate {trace_label} sorting choosers because not monotonic_increasing"
        )
        choosers = choosers.sort_index()

    if estimator:
        estimator.write_choosers(choosers)

    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]
    origin_col_name = model_settings["CHOOSER_ORIG_COL_NAME"]

    # alternatives are pre-sampled and annotated with logsums and pick_count
    # but we have to merge size_terms column into alt sample list
    destination_sample["size_term"] = reindex(
        destination_size_terms.size_term, destination_sample[alt_dest_col_name]
    )

    tracing.dump_df(DUMP, destination_sample, trace_label, "alternatives")

    constants = config.get_model_constants(model_settings)

    logger.info("Running tour_destination_simulate with %d persons", len(choosers))

    # create wrapper with keys for this lookup - in this case there is a home_zone_id in the choosers
    # and a zone_id in the alternatives which get merged during interaction
    # the skims will be available under the name "skims" for any @ expressions
    skim_dict = network_los.get_default_skim_dict()
    skims = skim_dict.wrap(origin_col_name, alt_dest_col_name)

    locals_d = {
        "skims": skims,
    }
    if constants is not None:
        locals_d.update(constants)

    tracing.dump_df(DUMP, choosers, trace_label, "choosers")

    log_alt_losers = config.setting("log_alt_losers", False)

    choices = interaction_sample_simulate(
        choosers,
        destination_sample,
        spec=model_spec,
        choice_column=alt_dest_col_name,
        log_alt_losers=log_alt_losers,
        want_logsums=want_logsums,
        skims=skims,
        locals_d=locals_d,
        chunk_size=chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
        trace_choice_name="destination",
        estimator=estimator,
    )

    if not want_logsums:
        # for consistency, always return a dataframe with canonical column name
        assert isinstance(choices, pd.Series)
        choices = choices.to_frame("choice")

    return choices


def run_tour_destination(
    tours,
    persons_merged,
    want_logsums,
    want_sample_table,
    model_settings,
    network_los,
    estimator,
    chunk_size,
    trace_hh_id,
    trace_label,
):

    size_term_calculator = SizeTermCalculator(model_settings["SIZE_TERM_SELECTOR"])

    # maps segment names to compact (integer) ids
    segments = model_settings["SEGMENTS"]

    chooser_segment_column = model_settings.get("CHOOSER_SEGMENT_COLUMN_NAME", None)
    if chooser_segment_column is None:
        assert (
            len(segments) == 1
        ), f"CHOOSER_SEGMENT_COLUMN_NAME not specified in model_settings to slice SEGMENTS: {segments}"

    choices_list = []
    sample_list = []
    for segment_name in segments:

        segment_trace_label = tracing.extend_trace_label(trace_label, segment_name)

        if chooser_segment_column is not None:
            choosers = tours[tours[chooser_segment_column] == segment_name]
        else:
            choosers = tours.copy()

        # Note: size_term_calculator omits zones with impossible alternatives (where dest size term is zero)
        segment_destination_size_terms = size_term_calculator.dest_size_terms_df(
            segment_name, segment_trace_label
        )

        if choosers.shape[0] == 0:
            logger.info(
                "%s skipping segment %s: no choosers", trace_label, segment_name
            )
            continue

        # - destination_sample
        spec_segment_name = segment_name  # spec_segment_name is segment_name
        location_sample_df = run_destination_sample(
            spec_segment_name,
            choosers,
            persons_merged,
            model_settings,
            network_los,
            segment_destination_size_terms,
            estimator,
            chunk_size=chunk_size,
            trace_label=tracing.extend_trace_label(segment_trace_label, "sample"),
        )

        # - destination_logsums
        tour_purpose = segment_name  # tour_purpose is segment_name
        location_sample_df = run_destination_logsums(
            tour_purpose,
            persons_merged,
            location_sample_df,
            model_settings,
            network_los,
            chunk_size=chunk_size,
            trace_label=tracing.extend_trace_label(segment_trace_label, "logsums"),
        )

        # - destination_simulate
        spec_segment_name = segment_name  # spec_segment_name is segment_name
        choices = run_destination_simulate(
            spec_segment_name,
            choosers,
            persons_merged,
            destination_sample=location_sample_df,
            want_logsums=want_logsums,
            model_settings=model_settings,
            network_los=network_los,
            destination_size_terms=segment_destination_size_terms,
            estimator=estimator,
            chunk_size=chunk_size,
            trace_label=tracing.extend_trace_label(segment_trace_label, "simulate"),
        )

        choices_list.append(choices)

        if want_sample_table:
            # FIXME - sample_table
            location_sample_df.set_index(
                model_settings["ALT_DEST_COL_NAME"], append=True, inplace=True
            )
            sample_list.append(location_sample_df)
        else:
            # del this so we dont hold active reference to it while run_location_sample is creating its replacement
            del location_sample_df

    if len(choices_list) > 0:
        choices_df = pd.concat(choices_list)
    else:
        # this will only happen with small samples (e.g. singleton) with no (e.g.) school segs
        logger.warning("%s no choices", trace_label)
        choices_df = pd.DataFrame(columns=["choice", "logsum"])

    if len(sample_list) > 0:
        save_sample_df = pd.concat(sample_list)
    else:
        # this could happen either with small samples as above, or if no saved sample desired
        save_sample_df = None

    return choices_df, save_sample_df



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import config, expressions, inject, pipeline, simulate, tracing

from .util import estimation
from .util.overlap import hh_time_window_overlap
from .util.tour_frequency import process_joint_tours

logger = logging.getLogger(__name__)


@inject.step()
def joint_tour_frequency(households, persons, chunk_size, trace_hh_id):
    """
    This model predicts the frequency of making fully joint trips (see the
    alternatives above).
    """
    trace_label = "joint_tour_frequency"
    model_settings_file_name = "joint_tour_frequency.yaml"

    estimator = estimation.manager.begin_estimation("joint_tour_frequency")

    model_settings = config.read_model_settings(model_settings_file_name)

    alternatives = simulate.read_model_alts(
        "joint_tour_frequency_alternatives.csv", set_index="alt"
    )

    # - only interested in households with more than one cdap travel_active person and
    # - at least one non-preschooler
    households = households.to_frame()
    multi_person_households = households[households.participates_in_jtf_model].copy()

    # - only interested in persons in multi_person_households
    # FIXME - gratuitous pathological efficiency move, just let yaml specify persons?
    persons = persons.to_frame()
    persons = persons[persons.household_id.isin(multi_person_households.index)]

    logger.info(
        "Running joint_tour_frequency with %d multi-person households"
        % multi_person_households.shape[0]
    )

    # - preprocessor
    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:

        locals_dict = {
            "persons": persons,
            "hh_time_window_overlap": hh_time_window_overlap,
        }

        expressions.assign_columns(
            df=multi_person_households,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label,
        )

    model_spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

    nest_spec = config.get_logit_model_settings(model_settings)
    constants = config.get_model_constants(model_settings)

    if estimator:
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_coefficients(coefficients_df, model_settings)
        estimator.write_choosers(multi_person_households)

    choices = simulate.simple_simulate(
        choosers=multi_person_households,
        spec=model_spec,
        nest_spec=nest_spec,
        locals_d=constants,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name="joint_tour_frequency",
        estimator=estimator,
    )

    # convert indexes to alternative names
    choices = pd.Series(model_spec.columns[choices.values], index=choices.index)

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(
            choices, "households", "joint_tour_frequency"
        )
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    # - create joint_tours based on joint_tour_frequency choices

    # - we need a person_id in order to generate the tour index (and for register_traceable_table)
    # - but we don't know the tour participants yet
    # - so we arbitrarily choose the first person in the household
    # - to be point person for the purpose of generating an index and setting origin
    temp_point_persons = persons.loc[persons.PNUM == 1]
    temp_point_persons["person_id"] = temp_point_persons.index
    temp_point_persons = temp_point_persons.set_index("household_id")
    temp_point_persons = temp_point_persons[["person_id", "home_zone_id"]]

    joint_tours = process_joint_tours(choices, alternatives, temp_point_persons)

    tours = pipeline.extend_table("tours", joint_tours)

    tracing.register_traceable_table("tours", joint_tours)
    pipeline.get_rn_generator().add_channel("tours", joint_tours)

    # - annotate households

    # we expect there to be an alt with no tours - which we can use to backfill non-travelers
    no_tours_alt = (alternatives.sum(axis=1) == 0).index[0]
    households["joint_tour_frequency"] = (
        choices.reindex(households.index).fillna(no_tours_alt).astype(str)
    )

    households["num_hh_joint_tours"] = (
        joint_tours.groupby("household_id")
        .size()
        .reindex(households.index)
        .fillna(0)
        .astype(np.int8)
    )

    pipeline.replace_table("households", households)

    tracing.print_summary(
        "joint_tour_frequency", households.joint_tour_frequency, value_counts=True
    )

    if trace_hh_id:
        tracing.trace_df(households, label="joint_tour_frequency.households")

        tracing.trace_df(
            joint_tours, label="joint_tour_frequency.joint_tours", slicer="household_id"
        )

    if estimator:
        survey_tours = estimation.manager.get_survey_table("tours")
        survey_tours = survey_tours[survey_tours.tour_category == "joint"]

        print(f"len(survey_tours) {len(survey_tours)}")
        print(f"len(joint_tours) {len(joint_tours)}")

        different = False
        survey_tours_not_in_tours = survey_tours[
            ~survey_tours.index.isin(joint_tours.index)
        ]
        if len(survey_tours_not_in_tours) > 0:
            print(f"survey_tours_not_in_tours\n{survey_tours_not_in_tours}")
            different = True
        tours_not_in_survey_tours = joint_tours[
            ~joint_tours.index.isin(survey_tours.index)
        ]
        if len(survey_tours_not_in_tours) > 0:
            print(f"tours_not_in_survey_tours\n{tours_not_in_survey_tours}")
            different = True
        assert not different



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import config, inject

logger = logging.getLogger(__name__)


@inject.injectable(cache=True)
def size_terms():
    f = config.config_file_path("destination_choice_size_terms.csv")
    return pd.read_csv(f, comment="#", index_col="segment")


def size_term(land_use, destination_choice_coefficients):
    """
    This method takes the land use data and multiplies various columns of the
    land use data by coefficients from the spec table in order
    to yield a size term (a linear combination of land use variables).

    Parameters
    ----------
    land_use : DataFrame
        A dataframe of land use attributes - the column names should match
        the index of destination_choice_coefficients
    destination_choice_coefficients : Series
        A series of coefficients for the land use attributes - the index
        describes the link to the land use table, and the values are floating
        points numbers used to do the linear combination

    Returns
    -------
    values : Series
        The index will be the same as land use, and the values will the
        linear combination of the land use table columns specified by the
        coefficients series.
    """
    coefficients = destination_choice_coefficients

    # first check for missing column in the land_use table
    missing = coefficients[~coefficients.index.isin(land_use.columns)]

    if len(missing) > 0:
        logger.warning("%s  missing columns in land use" % len(missing.index))
        for v in missing.index.values:
            logger.warning("missing: %s" % v)

    return land_use[coefficients.index].dot(coefficients)


def tour_destination_size_terms(land_use, size_terms, model_selector):
    """

     Parameters
     ----------
     land_use - pipeline table
     size_terms - pipeline table
     model_selector - str

     Returns
     -------

    ::

      pandas.dataframe
         one column per model_selector segment with index of land_use
         e.g. for model_selector 'workplace', columns will be work_low, work_med, ...
         and for model_selector 'trip', columns will be eatout, escort, othdiscr, ...

                  work_low    work_med  work_high   work_veryhigh
         zone_id                                         ...
         1      1267.00000     522.000  1108.000  1540.0000 ...
         2      1991.00000     824.500  1759.000  2420.0000 ...
     ...
    """

    land_use = land_use.to_frame()

    # don't count on land_use being sorted by index
    if not land_use.index.is_monotonic_increasing:
        land_use = land_use.sort_index()

    size_terms = size_terms[size_terms.model_selector == model_selector].copy()
    del size_terms["model_selector"]

    df = pd.DataFrame(
        {key: size_term(land_use, row) for key, row in size_terms.iterrows()},
        index=land_use.index,
    )

    assert land_use.index.name == "zone_id"
    df.index.name = land_use.index.name

    if not (df.dtypes == "float64").all():
        logger.warning("Surprised to find that not all size_terms were float64!")

    if df.isna().any(axis=None):
        logger.warning(
            f"tour_destination_size_terms with NAN values\n{df[df.isna().any(axis=1)]}"
        )
        assert not df.isna().any(axis=None)

    return df



import argparse
import os

import numpy as np
import openmatrix as omx
import pandas as pd

segments = {
    "test": "../data_2/taz.csv",  # crop to match 2 zone TAZs, includes univ
    "full": (0, 100000),
}

parser = argparse.ArgumentParser(description="crop SANDAG 1 zone raw_data")
parser.add_argument(
    "segment_name",
    metavar="segment_name",
    type=str,
    nargs=1,
    help=f"geography segmentation (e.g. full)",
)

parser.add_argument(
    "-c",
    "--check_geography",
    default=False,
    action="store_true",
    help="check consistency of TAZ zone_ids and foreign keys & write orphan_households file",
)

args = parser.parse_args()


segment_name = args.segment_name[0]
check_geography = args.check_geography

assert segment_name in segments.keys(), f"Unknown seg: {segment_name}"
if isinstance(segments[segment_name], str):
    zone_df = pd.read_csv(segments[segment_name])
    zones = zone_df["TAZ"].values
else:
    zone_min, zone_max = segments[segment_name]
    zones = range(zone_min, zone_max + 1)

input_dir = "./data_raw"
output_dir = f"./data_{segment_name}_1"


print(f"check_geography {check_geography}")

if not os.path.isdir(output_dir):
    print(f"creating output directory {output_dir}")
    os.mkdir(output_dir)


def input_path(file_name):
    return os.path.join(input_dir, file_name)


def output_path(file_name):
    return os.path.join(output_dir, file_name)


def integerize_id_columns(df, table_name):
    columns = ["TAZ", "household_id", "HHID", "taz"]
    for c in df.columns:
        if c in columns:
            print(f"converting {table_name}.{c} to int")
            if df[c].isnull().any():
                print(df[c][df[c].isnull()])
            df[c] = df[c].astype(int)


def read_csv(file_name, integerize=True):
    df = pd.read_csv(input_path(file_name))

    print(f"read {file_name} {df.shape}")

    return df


def to_csv(df, file_name):
    print(f"writing {file_name} {df.shape} {output_path(file_name)}")
    df.to_csv(output_path(file_name), index=False)


if check_geography:

    # ######## check for orphan_households not in any taz in land_use
    land_use = read_csv("land_use.csv")

    households = read_csv("households.csv")
    orphan_households = households[~households.TAZ.isin(land_use.TAZ)]
    print(f"{len(orphan_households)} orphan_households")

    # write orphan_households to INPUT directory (since it doesn't belong in output)
    file_name = "orphan_households.csv"
    print(f"writing {file_name} {orphan_households.shape} to {input_path(file_name)}")
    orphan_households.to_csv(input_path(file_name), index=False)


#
# land_use
#
land_use = read_csv("land_use.csv")
land_use = land_use[land_use["TAZ"].isin(zones)]
integerize_id_columns(land_use, "land_use")
land_use = land_use.sort_values("TAZ")

# move index col to front
land_use.insert(0, "TAZ", land_use.pop("TAZ"))

to_csv(land_use, "land_use.csv")

#
# households
#
households = read_csv("households.csv")
households = households[households["TAZ"].isin(land_use.TAZ)]
integerize_id_columns(households, "households")

to_csv(households, "households.csv")

#
# persons
#
persons = read_csv("persons.csv")
persons = persons[persons["household_id"].isin(households.HHID)]
integerize_id_columns(persons, "persons")

to_csv(persons, "persons.csv")

#
# skims
#
omx_infile_name = "skims.omx"
skim_data_type = np.float32

omx_in = omx.open_file(input_path(omx_infile_name))
print(f"omx_in shape {omx_in.shape()}")


zone = land_use.sort_values("TAZ")[["TAZ"]]
zone.index = zone.TAZ - 1
zone_indexes = zone.index.tolist()  # index of TAZ in skim (zero-based, no mapping)
zone_labels = zone.TAZ.tolist()  # TAZ in omx index order


# create
num_outfiles = 6 if segment_name == "full" else 1
if num_outfiles == 1:
    omx_out = [omx.open_file(output_path(f"skims1.omx"), "w")]
else:
    omx_out = [
        omx.open_file(output_path(f"skims{i+1}.omx"), "w") for i in range(num_outfiles)
    ]

for omx_file in omx_out:
    omx_file.create_mapping("ZONE", zone_labels)

iskim = 0
for mat_name in omx_in.list_matrices():

    # make sure we have a vanilla numpy array, not a CArray
    m = np.asanyarray(omx_in[mat_name]).astype(skim_data_type)
    m = m[zone_indexes, :][:, zone_indexes]
    print(f"{mat_name} {m.shape}")

    omx_file = omx_out[iskim % num_outfiles]
    omx_file[mat_name] = m
    iskim += 1


omx_in.close()
for omx_file in omx_out:
    omx_file.close()



# ActivitySim
# See full license in LICENSE.txt.
import logging
import sys

import numpy as np
import pandas as pd

from activitysim.core import config, inject, pipeline
from activitysim.core.config import setting

logger = logging.getLogger(__name__)


def track_skim_usage(output_dir):
    """
    write statistics on skim usage (diagnostic to detect loading of un-needed skims)

    FIXME - have not yet implemented a facility to avoid loading of unused skims

    FIXME - if resume_after, this will only reflect skims used after resume

    Parameters
    ----------
    output_dir: str

    """
    pd.options.display.max_columns = 500
    pd.options.display.max_rows = 100

    skim_dict = inject.get_injectable("skim_dict")

    mode = "wb" if sys.version_info < (3,) else "w"
    with open(config.output_file_path("skim_usage.txt"), mode) as output_file:

        print("\n### skim_dict usage", file=output_file)
        for key in skim_dict.get_skim_usage():
            print(key, file=output_file)

        unused = set(k for k in skim_dict.skim_info.base_keys) - set(
            k for k in skim_dict.get_skim_usage()
        )

        for key in unused:
            print(key, file=output_file)


def previous_write_data_dictionary(output_dir):
    """
    Write table_name, number of rows, columns, and bytes for each checkpointed table

    Parameters
    ----------
    output_dir: str

    """

    model_settings = config.read_model_settings("write_data_dictionary")
    txt_format = model_settings.get("txt_format", "data_dict.txt")
    csv_format = model_settings.get("csv_format", "data_dict.csv")

    if txt_format:

        output_file_path = config.output_file_path(txt_format)

        pd.options.display.max_columns = 500
        pd.options.display.max_rows = 100

        output_tables = pipeline.checkpointed_tables()

        # write data dictionary for all checkpointed_tables

        with open(output_file_path, "w") as output_file:
            for table_name in output_tables:
                df = inject.get_table(table_name, None).to_frame()

                print("\n### %s %s" % (table_name, df.shape), file=output_file)
                print("index:", df.index.name, df.index.dtype, file=output_file)
                print(df.dtypes, file=output_file)


def write_data_dictionary(output_dir):
    """
    Write table schema for all tables

    model settings
        txt_format: output text file name (default data_dict.txt) or empty to suppress txt output
        csv_format: output csv file name (default data_dict.tcsvxt) or empty to suppress txt output

        schema_tables: list of tables to include in output (defaults to all checkpointed tables)

    for each table, write column names, dtype, and checkpoint added)

    text format writes individual table schemas to a single text file
    csv format writes all tables together with an additional table_name column

    Parameters
    ----------
    output_dir: str

    """

    model_settings = config.read_model_settings("write_data_dictionary")
    txt_format = model_settings.get("txt_format", "data_dict.txt")
    csv_format = model_settings.get("csv_format", "data_dict.csv")

    if not (csv_format or txt_format):
        logger.warning(
            f"write_data_dictionary step invoked but neither 'txt_format' nor 'csv_format' specified"
        )
        return

    table_names = pipeline.registered_tables()

    # use table_names list from model_settings, if provided
    schema_tables = model_settings.get("tables", None)
    if schema_tables:
        table_names = [c for c in schema_tables if c in table_names]

    # initialize schema as dict of dataframe[table_name, column_name, dtype, checkpoint]
    schema = dict()
    final_shapes = dict()
    for table_name in table_names:
        df = pipeline.get_table(table_name)

        final_shapes[table_name] = df.shape

        if df.index.name and df.index.name not in df.columns:
            df = df.reset_index()
        info = (
            df.dtypes.astype(str)
            .to_frame("dtype")
            .reset_index()
            .rename(columns={"index": "column_name"})
        )
        info["checkpoint"] = ""

        info.insert(loc=0, column="table_name", value=table_name)
        schema[table_name] = info

    # annotate schema.info with name of checkpoint columns were first seen
    for _, row in pipeline.get_checkpoints().iterrows():

        checkpoint_name = row[pipeline.CHECKPOINT_NAME]

        for table_name in table_names:

            # no change to table in this checkpoint
            if row.get(table_name, None) != checkpoint_name:
                continue

            # get the checkpointed version of the table
            df = pipeline.get_table(table_name, checkpoint_name)

            if df.index.name and df.index.name not in df.columns:
                df = df.reset_index()

            info = schema.get(table_name, None)

            # tag any new columns with checkpoint name
            prev_columns = info[info.checkpoint != ""].column_name.values
            new_cols = [c for c in df.columns.values if c not in prev_columns]
            is_new_column_this_checkpoont = info.column_name.isin(new_cols)
            info.checkpoint = np.where(
                is_new_column_this_checkpoont, checkpoint_name, info.checkpoint
            )

            schema[table_name] = info

    schema_df = pd.concat(schema.values())

    if csv_format:
        schema_df.to_csv(config.output_file_path(csv_format), header=True, index=False)

    if txt_format:
        with open(config.output_file_path(txt_format), "w") as output_file:

            # get max schema column widths from omnibus table
            col_width = {c: schema_df[c].str.len().max() + 2 for c in schema_df}

            for table_name in table_names:
                info = schema.get(table_name, None)

                columns_to_print = ["column_name", "dtype", "checkpoint"]
                info = info[columns_to_print].copy()

                # normalize schema columns widths across all table schemas for unified output formatting
                for c in info:
                    info[c] = info[c].str.pad(col_width[c], side="right")
                info.columns = [c.ljust(col_width[c]) for c in info.columns]

                info = info.to_string(index=False)

                print(
                    f"###\n### {table_name} {final_shapes[table_name]}\n###\n",
                    file=output_file,
                )
                print(f"{info}\n", file=output_file)


def write_tables(output_dir):
    """
    Write pipeline tables as csv files (in output directory) as specified by output_tables list
    in settings file.

    'output_tables' can specify either a list of output tables to include or to skip
    if no output_tables list is specified, then all checkpointed tables will be written

    To write all output tables EXCEPT the households and persons tables:

    ::

      output_tables:
        action: skip
        tables:
          - households
          - persons

    To write ONLY the households table:

    ::

      output_tables:
        action: include
        tables:
           - households

    To write tables into a single HDF5 store instead of individual CSVs, use the h5_store flag:

    ::

      output_tables:
        h5_store: True
        action: include
        tables:
           - households

    Parameters
    ----------
    output_dir: str

    """

    output_tables_settings_name = "output_tables"

    output_tables_settings = setting(output_tables_settings_name)

    if output_tables_settings is None:
        logger.info("No output_tables specified in settings file. Nothing to write.")
        return

    action = output_tables_settings.get("action")
    tables = output_tables_settings.get("tables")
    prefix = output_tables_settings.get("prefix", "final_")
    h5_store = output_tables_settings.get("h5_store", False)
    sort = output_tables_settings.get("sort", False)

    registered_tables = pipeline.registered_tables()
    if action == "include":
        # interpret empty or missing tables setting to mean include all registered tables
        output_tables_list = tables if tables is not None else registered_tables
    elif action == "skip":
        output_tables_list = [t for t in registered_tables if t not in tables]
    else:
        raise "expected %s action '%s' to be either 'include' or 'skip'" % (
            output_tables_settings_name,
            action,
        )

    for table_name in output_tables_list:

        if table_name == "checkpoints":
            df = pipeline.get_checkpoints()
        else:
            if table_name not in registered_tables:
                logger.warning("Skipping '%s': Table not found." % table_name)
                continue
            df = pipeline.get_table(table_name)

            if sort:
                traceable_table_indexes = inject.get_injectable(
                    "traceable_table_indexes", {}
                )

                if df.index.name in traceable_table_indexes:
                    df = df.sort_index()
                    logger.debug(
                        f"write_tables sorting {table_name} on index {df.index.name}"
                    )
                else:
                    # find all registered columns we can use to sort this table
                    # (they are ordered appropriately in traceable_table_indexes)
                    sort_columns = [
                        c for c in traceable_table_indexes if c in df.columns
                    ]
                    if len(sort_columns) > 0:
                        df = df.sort_values(by=sort_columns)
                        logger.debug(
                            f"write_tables sorting {table_name} on columns {sort_columns}"
                        )
                    else:
                        logger.debug(
                            f"write_tables sorting {table_name} on unrecognized index {df.index.name}"
                        )
                        df = df.sort_index()

        if h5_store:
            file_path = config.output_file_path("%soutput_tables.h5" % prefix)
            df.to_hdf(file_path, key=table_name, mode="a", format="fixed")
        else:
            file_name = "%s%s.csv" % (prefix, table_name)
            file_path = config.output_file_path(file_name)

            # include the index if it has a name or is a MultiIndex
            write_index = df.index.name is not None or isinstance(
                df.index, pd.MultiIndex
            )

            df.to_csv(file_path, index=write_index)



import itertools
import logging
import os
import re
from pathlib import Path
from typing import Mapping

import numpy as np
import pandas as pd
import yaml
from larch import DataFrames, Model, P, X
from larch.log import logger_name
from larch.util import Dict

from .general import (
    apply_coefficients,
    cv_to_ca,
    linear_utility_from_spec,
    remove_apostrophes,
)

_logger = logging.getLogger(logger_name)


def interaction_simulate_data(
    name="non_mandatory_tour_frequency",
    edb_directory="output/estimation_data_bundle/{name}/",
    settings_file="{name}_model_settings.yaml",
    spec_file="{name}_SPEC.csv",
    alt_def_file="{name}_alternatives.csv",
    coefficients_files="{segment_name}/{name}_coefficients_{segment_name}.csv",
    chooser_data_files="{segment_name}/{name}_choosers_combined.csv",
    alt_values_files="{segment_name}/{name}_interaction_expression_values.csv",
):
    edb_directory = edb_directory.format(name=name)

    def _read_csv(filename, **kwargs):
        filename = filename.format(name=name)
        return pd.read_csv(os.path.join(edb_directory, filename), **kwargs)

    settings_file = settings_file.format(name=name)
    with open(os.path.join(edb_directory, settings_file), "r") as yf:
        settings = yaml.load(
            yf,
            Loader=yaml.SafeLoader,
        )

    coefficients = {}
    chooser_data = {}
    alt_values = {}

    segment_names = [s["NAME"] for s in settings["SPEC_SEGMENTS"]]

    for segment_name in segment_names:
        coefficients[segment_name] = _read_csv(
            coefficients_files.format(name=name, segment_name=segment_name),
            index_col="coefficient_name",
        )
        chooser_data[segment_name] = _read_csv(
            chooser_data_files.format(name=name, segment_name=segment_name),
        )
        alt_values[segment_name] = _read_csv(
            alt_values_files.format(name=name, segment_name=segment_name),
        )

    spec = _read_csv(
        spec_file,
    )
    spec = remove_apostrophes(spec, ["Label"])
    # alt_names = list(spec.columns[3:])
    # alt_codes = np.arange(1, len(alt_names) + 1)
    # alt_names_to_codes = dict(zip(alt_names, alt_codes))
    # alt_codes_to_names = dict(zip(alt_codes, alt_names))

    alt_def = _read_csv(alt_def_file.format(name=name), index_col=0)

    return Dict(
        edb_directory=Path(edb_directory),
        settings=settings,
        chooser_data=chooser_data,
        coefficients=coefficients,
        alt_values=alt_values,
        spec=spec,
        alt_def=alt_def,
    )


def link_same_value_coefficients(segment_names, coefficients, spec):
    # Assume all coefficients with exactly equal current values are
    # actually the same estimated coefficient value and should be
    # treated as such by Larch.  Comment out this function where called to relax
    # this assumption, although be careful about the number of unique
    # parameters to estimate in these models.

    relabel_coef = {}
    for segment_name in segment_names:
        coef_backwards_map = dict(
            [(j, i) for i, j in coefficients[segment_name]["value"].items()]
        )
        relabel_coef[segment_name] = r = coefficients[segment_name]["value"].map(
            coef_backwards_map
        )
        spec[segment_name] = spec[segment_name].map(r)
    return relabel_coef


def unavail_parameters(model):
    return model.pf.index[(model.pf.value < -900) & (model.pf.holdfast != 0)]


def unavail_data_cols(model):
    locks = unavail_parameters(model)
    return [i.data for i in model.utility_ca if i.param in locks]


def unavail(model, x_ca):
    lock_data = unavail_data_cols(model)
    if len(lock_data):
        unav = x_ca[lock_data[0]] > 0
        for j in lock_data[1:]:
            unav |= x_ca[j] > 0
    return unav


def nonmand_tour_freq_model(
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    data = interaction_simulate_data(
        name="non_mandatory_tour_frequency",
        edb_directory=edb_directory,
    )

    settings = data.settings
    segment_names = [s["NAME"] for s in settings["SPEC_SEGMENTS"]]
    data.relabel_coef = link_same_value_coefficients(
        segment_names, data.coefficients, data.spec
    )
    spec = data.spec
    coefficients = data.coefficients
    chooser_data = data.chooser_data
    alt_values = data.alt_values
    alt_def = data.alt_def

    m = {}
    for segment_name in segment_names:
        segment_model = m[segment_name] = Model()
        # One of the alternatives is coded as 0, so
        # we need to explicitly initialize the MNL nesting graph
        # and set to root_id to a value other than zero.
        segment_model.initialize_graph(alternative_codes=alt_def.index, root_id=9999)

        # Utility specifications
        segment_model.utility_ca = linear_utility_from_spec(
            spec,
            x_col="Label",
            p_col=segment_name,
        )
        apply_coefficients(coefficients[segment_name], segment_model)
        segment_model.choice_co_code = "override_choice"

        # Attach Data
        x_co = (
            chooser_data[segment_name]
            .set_index("person_id")
            .rename(columns={"TAZ": "HOMETAZ"})
        )
        x_ca = cv_to_ca(alt_values[segment_name].set_index(["person_id", "variable"]))
        d = DataFrames(
            co=x_co,
            ca=x_ca,
            av=~unavail(segment_model, x_ca),
        )
        m[segment_name].dataservice = d

    if return_data:
        return m, data
    return m



from activitysim.benchmarking.componentwise import (
    template_component_timings,
    template_setup_cache,
)

from .sandag_full import *

EXAMPLE_NAME = "placeholder_sandag_2_zone_full"
CONFIGS_DIRS = ("configs_benchmarking", "configs_2_zone", "placeholder_psrc/configs")
DYNAMIC_CONFIG_DIR = "bench_configs"
DATA_DIR = "data_2"
OUTPUT_DIR = "output_2"
VERSION = "1"


def setup_cache():
    template_setup_cache(
        EXAMPLE_NAME,
        COMPONENT_NAMES,
        BENCHMARK_SETTINGS,
        dict(
            read_skim_cache=SKIM_CACHE,
            write_skim_cache=SKIM_CACHE,
        ),
        CONFIGS_DIRS,
        DATA_DIR,
        OUTPUT_DIR,
    )


template_component_timings(
    globals(),
    COMPONENT_NAMES,
    EXAMPLE_NAME,
    (DYNAMIC_CONFIG_DIR, *CONFIGS_DIRS),
    DATA_DIR,
    OUTPUT_DIR,
    PRELOAD_INJECTABLES,
    REPEAT,
    NUMBER,
    TIMEOUT,
    VERSION,
)



# ActivitySim
# See full license in LICENSE.txt.

import logging

import numpy as np
import pandas as pd

from activitysim.core import (
    assign,
    chunk,
    config,
    expressions,
    inject,
    los,
    pipeline,
    simulate,
    tracing,
)
from activitysim.core.pathbuilder import TransitVirtualPathBuilder
from activitysim.core.util import assign_in_place

from .util import estimation
from .util.mode import mode_choice_simulate

logger = logging.getLogger(__name__)


@inject.step()
def trip_mode_choice(trips, network_los, chunk_size, trace_hh_id):
    """
    Trip mode choice - compute trip_mode (same values as for tour_mode) for each trip.

    Modes for each primary tour putpose are calculated separately because they have different
    coefficient values (stored in trip_mode_choice_coefficients.csv coefficient file.)

    Adds trip_mode column to trip table
    """

    trace_label = "trip_mode_choice"
    model_settings_file_name = "trip_mode_choice.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    logsum_column_name = model_settings.get("MODE_CHOICE_LOGSUM_COLUMN_NAME")
    mode_column_name = "trip_mode"

    trips_df = trips.to_frame()
    logger.info("Running %s with %d trips", trace_label, trips_df.shape[0])

    # give trip mode choice the option to run without calling tours_merged. Useful for xborder
    # model where tour_od_choice needs trip mode choice logsums before some of the join keys
    # needed by tour_merged (e.g. home_zone_id) exist
    tours_cols = [
        col
        for col in model_settings["TOURS_MERGED_CHOOSER_COLUMNS"]
        if col not in trips_df.columns
    ]
    if len(tours_cols) > 0:
        tours_merged = inject.get_table("tours_merged").to_frame(columns=tours_cols)
    else:
        tours_merged = pd.DataFrame()

    # - trips_merged - merge trips and tours_merged
    trips_merged = pd.merge(
        trips_df, tours_merged, left_on="tour_id", right_index=True, how="left"
    )
    assert trips_merged.index.equals(trips.index)

    tracing.print_summary(
        "primary_purpose", trips_df.primary_purpose, value_counts=True
    )

    # setup skim keys
    assert "trip_period" not in trips_merged
    trips_merged["trip_period"] = network_los.skim_time_period_label(
        trips_merged.depart
    )

    orig_col = "origin"
    dest_col = "destination"
    min_per_period = network_los.skim_time_periods["period_minutes"]
    periods_per_hour = 60 / min_per_period

    constants = {}
    constants.update(config.get_model_constants(model_settings))
    constants.update(
        {
            "ORIGIN": orig_col,
            "DESTINATION": dest_col,
            "MIN_PER_PERIOD": min_per_period,
            "PERIODS_PER_HOUR": periods_per_hour,
        }
    )

    skim_dict = network_los.get_default_skim_dict()

    odt_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=orig_col, dest_key=dest_col, dim3_key="trip_period"
    )
    dot_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=dest_col, dest_key=orig_col, dim3_key="trip_period"
    )
    od_skim_wrapper = skim_dict.wrap("origin", "destination")

    skims = {
        "odt_skims": odt_skim_stack_wrapper,
        "dot_skims": dot_skim_stack_wrapper,
        "od_skims": od_skim_wrapper,
    }

    if network_los.zone_system == los.THREE_ZONE:
        # fixme - is this a lightweight object?
        tvpb = network_los.tvpb
        tvpb_recipe = model_settings.get("TVPB_recipe", "tour_mode_choice")
        tvpb_logsum_odt = tvpb.wrap_logsum(
            orig_key=orig_col,
            dest_key=dest_col,
            tod_key="trip_period",
            segment_key="demographic_segment",
            recipe=tvpb_recipe,
            cache_choices=True,
            trace_label=trace_label,
            tag="tvpb_logsum_odt",
        )
        skims.update(
            {
                "tvpb_logsum_odt": tvpb_logsum_odt,
            }
        )

        # This if-clause gives the user the option of NOT inheriting constants
        # from the tvpb settings. previously, these constants were inherited
        # automatically, which had the undesirable effect of overwriting any
        # trip mode choice model constants/coefficients that shared the same
        # name. The default behavior is still the same (True), but the user
        # can now avoid any chance of squashing these local variables by
        # adding `use_TVPB_constants: False` to the trip_mode_choice.yaml file.
        # the tvpb will still use the constants as defined in the recipe
        # specified above in `tvpb.wrap_logsum()` but they will not be used
        # in the trip mode choice expressions.
        if model_settings.get("use_TVPB_constants", True):
            constants.update(
                network_los.setting("TVPB_SETTINGS.tour_mode_choice.CONSTANTS")
            )

    # don't create estimation data bundle if trip mode choice is being called
    # from another model step (e.g. tour mode choice logsum creation)
    if pipeline._PIPELINE.rng().step_name != "trip_mode_choice":
        estimator = None
    else:
        estimator = estimation.manager.begin_estimation("trip_mode_choice")
    if estimator:
        estimator.write_coefficients(model_settings=model_settings)
        estimator.write_coefficients_template(model_settings=model_settings)
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)

    model_spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
    nest_spec = config.get_logit_model_settings(model_settings)

    choices_list = []
    for primary_purpose, trips_segment in trips_merged.groupby("primary_purpose"):

        segment_trace_label = tracing.extend_trace_label(trace_label, primary_purpose)

        logger.info(
            "trip_mode_choice tour_type '%s' (%s trips)"
            % (
                primary_purpose,
                len(trips_segment.index),
            )
        )

        # name index so tracing knows how to slice
        assert trips_segment.index.name == "trip_id"

        if network_los.zone_system == los.THREE_ZONE:
            tvpb_logsum_odt.extend_trace_label(primary_purpose)
            # tvpb_logsum_dot.extend_trace_label(primary_purpose)

        coefficients = simulate.get_segment_coefficients(
            model_settings, primary_purpose
        )

        locals_dict = {}
        locals_dict.update(constants)

        constants_keys = constants.keys()
        if any([coeff in constants_keys for coeff in coefficients.keys()]):
            logger.warning("coefficients are obscuring constants in locals_dict")
        locals_dict.update(coefficients)

        # have to initialize chunker for preprocessing in order to access
        # tvpb logsum terms in preprocessor expressions.
        with chunk.chunk_log(
            tracing.extend_trace_label(trace_label, "preprocessing"), base=True
        ):
            expressions.annotate_preprocessors(
                trips_segment, locals_dict, skims, model_settings, segment_trace_label
            )

        if estimator:
            # write choosers after annotation
            estimator.write_choosers(trips_segment)

        locals_dict.update(skims)

        choices = mode_choice_simulate(
            choosers=trips_segment,
            spec=simulate.eval_coefficients(model_spec, coefficients, estimator),
            nest_spec=simulate.eval_nest_coefficients(
                nest_spec, coefficients, segment_trace_label
            ),
            skims=skims,
            locals_d=locals_dict,
            chunk_size=chunk_size,
            mode_column_name=mode_column_name,
            logsum_column_name=logsum_column_name,
            trace_label=segment_trace_label,
            trace_choice_name="trip_mode_choice",
            estimator=estimator,
        )

        if trace_hh_id:
            # trace the coefficients
            tracing.trace_df(
                pd.Series(locals_dict),
                label=tracing.extend_trace_label(segment_trace_label, "constants"),
                transpose=False,
                slicer="NONE",
            )

            # so we can trace with annotations
            assign_in_place(trips_segment, choices)

            tracing.trace_df(
                trips_segment,
                label=tracing.extend_trace_label(segment_trace_label, "trip_mode"),
                slicer="tour_id",
                index_label="tour_id",
                warn_if_empty=True,
            )

        choices_list.append(choices)

    choices_df = pd.concat(choices_list)

    # add cached tvpb_logsum tap choices for modes specified in tvpb_mode_path_types
    if network_los.zone_system == los.THREE_ZONE:

        tvpb_mode_path_types = model_settings.get("tvpb_mode_path_types")
        for mode, path_type in tvpb_mode_path_types.items():

            skim_cache = tvpb_logsum_odt.cache[path_type]

            for c in skim_cache:
                dest_col = c
                if dest_col not in choices_df:
                    choices_df[dest_col] = (
                        np.nan if pd.api.types.is_numeric_dtype(skim_cache[c]) else ""
                    )
                choices_df[dest_col].where(
                    choices_df[mode_column_name] != mode, skim_cache[c], inplace=True
                )

    if estimator:
        estimator.write_choices(choices_df.trip_mode)
        choices_df.trip_mode = estimator.get_survey_values(
            choices_df.trip_mode, "trips", "trip_mode"
        )
        estimator.write_override_choices(choices_df.trip_mode)
        estimator.end_estimation()
    trips_df = trips.to_frame()
    assign_in_place(trips_df, choices_df)

    tracing.print_summary("trip_modes", trips_merged.tour_mode, value_counts=True)

    tracing.print_summary(
        "trip_mode_choice choices", trips_df[mode_column_name], value_counts=True
    )

    assert not trips_df[mode_column_name].isnull().any()

    pipeline.replace_table("trips", trips_df)

    if trace_hh_id:
        tracing.trace_df(
            trips_df,
            label=tracing.extend_trace_label(trace_label, "trip_mode"),
            slicer="trip_id",
            index_label="trip_id",
            warn_if_empty=True,
        )



# ActivitySim
# See full license in LICENSE.txt.
import logging

from activitysim.core import config, expressions, los, simulate, tracing
from activitysim.core.pathbuilder import TransitVirtualPathBuilder

logger = logging.getLogger(__name__)


def filter_chooser_columns(choosers, logsum_settings, model_settings):

    chooser_columns = logsum_settings.get("LOGSUM_CHOOSER_COLUMNS", [])

    if (
        "CHOOSER_ORIG_COL_NAME" in model_settings
        and model_settings["CHOOSER_ORIG_COL_NAME"] not in chooser_columns
    ):
        chooser_columns.append(model_settings["CHOOSER_ORIG_COL_NAME"])

    missing_columns = [c for c in chooser_columns if c not in choosers]
    if missing_columns:
        logger.debug(
            "logsum.filter_chooser_columns missing_columns %s" % missing_columns
        )

    # ignore any columns not appearing in choosers df
    chooser_columns = [c for c in chooser_columns if c in choosers]

    choosers = choosers[chooser_columns]
    return choosers


def compute_logsums(
    choosers,
    tour_purpose,
    logsum_settings,
    model_settings,
    network_los,
    chunk_size,
    chunk_tag,
    trace_label,
    in_period_col=None,
    out_period_col=None,
    duration_col=None,
):
    """

    Parameters
    ----------
    choosers
    tour_purpose
    logsum_settings
    model_settings
    network_los
    chunk_size
    trace_hh_id
    trace_label

    Returns
    -------
    logsums: pandas series
        computed logsums with same index as choosers
    """

    trace_label = tracing.extend_trace_label(trace_label, "compute_logsums")
    logger.debug("Running compute_logsums with %d choosers" % choosers.shape[0])

    # compute_logsums needs to know name of dest column in interaction_sample
    orig_col_name = model_settings["CHOOSER_ORIG_COL_NAME"]
    dest_col_name = model_settings["ALT_DEST_COL_NAME"]

    # FIXME - are we ok with altering choosers (so caller doesn't have to set these)?
    if (in_period_col is not None) and (out_period_col is not None):
        choosers["in_period"] = network_los.skim_time_period_label(
            choosers[in_period_col]
        )
        choosers["out_period"] = network_los.skim_time_period_label(
            choosers[out_period_col]
        )
    elif ("in_period" not in choosers.columns) and (
        "out_period" not in choosers.columns
    ):
        if (
            type(model_settings["IN_PERIOD"]) is dict
            and type(model_settings["OUT_PERIOD"]) is dict
        ):
            if (
                tour_purpose in model_settings["IN_PERIOD"]
                and tour_purpose in model_settings["OUT_PERIOD"]
            ):
                choosers["in_period"] = network_los.skim_time_period_label(
                    model_settings["IN_PERIOD"][tour_purpose]
                )
                choosers["out_period"] = network_los.skim_time_period_label(
                    model_settings["OUT_PERIOD"][tour_purpose]
                )
        else:
            choosers["in_period"] = network_los.skim_time_period_label(
                model_settings["IN_PERIOD"]
            )
            choosers["out_period"] = network_los.skim_time_period_label(
                model_settings["OUT_PERIOD"]
            )
    else:
        logger.error("Choosers table already has columns 'in_period' and 'out_period'.")

    if duration_col is not None:
        choosers["duration"] = choosers[duration_col]
    elif "duration" not in choosers.columns:
        if (
            type(model_settings["IN_PERIOD"]) is dict
            and type(model_settings["OUT_PERIOD"]) is dict
        ):
            if (
                tour_purpose in model_settings["IN_PERIOD"]
                and tour_purpose in model_settings["OUT_PERIOD"]
            ):
                choosers["duration"] = (
                    model_settings["IN_PERIOD"][tour_purpose]
                    - model_settings["OUT_PERIOD"][tour_purpose]
                )
        else:
            choosers["duration"] = (
                model_settings["IN_PERIOD"] - model_settings["OUT_PERIOD"]
            )
    else:
        logger.error("Choosers table already has column 'duration'.")

    logsum_spec = simulate.read_model_spec(file_name=logsum_settings["SPEC"])
    coefficients = simulate.get_segment_coefficients(logsum_settings, tour_purpose)

    logsum_spec = simulate.eval_coefficients(logsum_spec, coefficients, estimator=None)

    nest_spec = config.get_logit_model_settings(logsum_settings)
    nest_spec = simulate.eval_nest_coefficients(nest_spec, coefficients, trace_label)

    locals_dict = {}
    # model_constants can appear in expressions
    locals_dict.update(config.get_model_constants(logsum_settings))
    # constrained coefficients can appear in expressions
    locals_dict.update(coefficients)

    # setup skim keys
    skim_dict = network_los.get_default_skim_dict()

    odt_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=orig_col_name, dest_key=dest_col_name, dim3_key="out_period"
    )
    dot_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=dest_col_name, dest_key=orig_col_name, dim3_key="in_period"
    )
    odr_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=orig_col_name, dest_key=dest_col_name, dim3_key="in_period"
    )
    dor_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=dest_col_name, dest_key=orig_col_name, dim3_key="out_period"
    )
    od_skim_stack_wrapper = skim_dict.wrap(orig_col_name, dest_col_name)

    skims = {
        "odt_skims": odt_skim_stack_wrapper,
        "dot_skims": dot_skim_stack_wrapper,
        "odr_skims": odr_skim_stack_wrapper,
        "dor_skims": dor_skim_stack_wrapper,
        "od_skims": od_skim_stack_wrapper,
        "orig_col_name": orig_col_name,
        "dest_col_name": dest_col_name,
    }

    if network_los.zone_system == los.THREE_ZONE:
        # fixme - is this a lightweight object?
        tvpb = network_los.tvpb

        tvpb_logsum_odt = tvpb.wrap_logsum(
            orig_key=orig_col_name,
            dest_key=dest_col_name,
            tod_key="out_period",
            segment_key="demographic_segment",
            trace_label=trace_label,
            tag="tvpb_logsum_odt",
        )
        tvpb_logsum_dot = tvpb.wrap_logsum(
            orig_key=dest_col_name,
            dest_key=orig_col_name,
            tod_key="in_period",
            segment_key="demographic_segment",
            trace_label=trace_label,
            tag="tvpb_logsum_dot",
        )

        skims.update(
            {"tvpb_logsum_odt": tvpb_logsum_odt, "tvpb_logsum_dot": tvpb_logsum_dot}
        )

        # TVPB constants can appear in expressions
        if logsum_settings.get("use_TVPB_constants", True):
            locals_dict.update(
                network_los.setting("TVPB_SETTINGS.tour_mode_choice.CONSTANTS")
            )

    locals_dict.update(skims)

    # - run preprocessor to annotate choosers
    # allow specification of alternate preprocessor for nontour choosers
    preprocessor = model_settings.get("LOGSUM_PREPROCESSOR", "preprocessor")
    preprocessor_settings = logsum_settings[preprocessor]

    if preprocessor_settings:

        simulate.set_skim_wrapper_targets(choosers, skims)

        expressions.assign_columns(
            df=choosers,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label,
        )

    logsums = simulate.simple_simulate_logsums(
        choosers,
        logsum_spec,
        nest_spec,
        skims=skims,
        locals_d=locals_dict,
        chunk_size=chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
    )

    return logsums



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import (
    chunk,
    config,
    expressions,
    inject,
    logit,
    pipeline,
    simulate,
    tracing,
)

from .util import estimation

logger = logging.getLogger(__name__)


PROBS_JOIN_COLUMNS = ["primary_purpose", "outbound", "person_type"]


def map_coefficients(spec, coefficients):
    if isinstance(coefficients, pd.DataFrame):
        assert "value" in coefficients.columns
        coefficients = coefficients["value"].to_dict()

    assert isinstance(
        coefficients, dict
    ), "map_coefficients doesn't grok type of coefficients: %s" % (type(coefficients))

    for c in spec.columns:
        if c == simulate.SPEC_LABEL_NAME:
            continue
        spec[c] = spec[c].map(coefficients).astype(np.float32)

    assert not spec.isnull().any()

    return spec


def choose_intermediate_trip_purpose(
    trips,
    probs_spec,
    estimator,
    probs_join_cols,
    use_depart_time,
    trace_hh_id,
    trace_label,
):
    """
    chose purpose for intermediate trips based on probs_spec
    which assigns relative weights (summing to 1) to the possible purpose choices

    Returns
    -------
    purpose: pandas.Series of purpose (str) indexed by trip_id
    """

    non_purpose_cols = probs_join_cols.copy()
    if use_depart_time:
        non_purpose_cols += ["depart_range_start", "depart_range_end"]
    purpose_cols = [c for c in probs_spec.columns if c not in non_purpose_cols]

    num_trips = len(trips.index)
    have_trace_targets = trace_hh_id and tracing.has_trace_targets(trips)

    # probs should sum to 1 across rows
    sum_probs = probs_spec[purpose_cols].sum(axis=1)
    probs_spec.loc[:, purpose_cols] = probs_spec.loc[:, purpose_cols].div(
        sum_probs, axis=0
    )

    # left join trips to probs (there may be multiple rows per trip for multiple depart ranges)
    choosers = pd.merge(
        trips.reset_index(), probs_spec, on=probs_join_cols, how="left"
    ).set_index("trip_id")
    chunk.log_df(trace_label, "choosers", choosers)

    if use_depart_time:

        # select the matching depart range (this should result on in exactly one chooser row per trip)
        chooser_probs = (choosers.start >= choosers["depart_range_start"]) & (
            choosers.start <= choosers["depart_range_end"]
        )

        # if we failed to match a row in probs_spec
        if chooser_probs.sum() < num_trips:

            # this can happen if the spec doesn't have probs for the trips matching a trip's probs_join_cols
            missing_trip_ids = trips.index[
                ~trips.index.isin(choosers.index[chooser_probs])
            ].values
            unmatched_choosers = choosers[choosers.index.isin(missing_trip_ids)]
            unmatched_choosers = unmatched_choosers[
                ["person_id", "start"] + non_purpose_cols
            ]

            # join to persons for better diagnostics
            persons = inject.get_table("persons").to_frame()
            persons_cols = [
                "age",
                "is_worker",
                "is_student",
                "is_gradeschool",
                "is_highschool",
                "is_university",
            ]
            unmatched_choosers = pd.merge(
                unmatched_choosers,
                persons[[col for col in persons_cols if col in persons.columns]],
                left_on="person_id",
                right_index=True,
                how="left",
            )

            file_name = "%s.UNMATCHED_PROBS" % trace_label
            logger.error(
                "%s %s of %s intermediate trips could not be matched to probs based on join columns  %s"
                % (trace_label, len(unmatched_choosers), len(choosers), probs_join_cols)
            )
            logger.info(
                "Writing %s unmatched choosers to %s"
                % (
                    len(unmatched_choosers),
                    file_name,
                )
            )
            tracing.write_csv(unmatched_choosers, file_name=file_name, transpose=False)
            raise RuntimeError(
                "Some trips could not be matched to probs based on join columns %s."
                % probs_join_cols
            )

        # select the matching depart range (this should result on in exactly one chooser row per trip)
        choosers = choosers[chooser_probs]

    # choosers should now match trips row for row
    assert choosers.index.identical(trips.index)

    if estimator:
        probs_cols = list(probs_spec.columns)
        print(choosers[probs_cols])
        estimator.write_table(choosers[probs_cols], "probs", append=True)

    choices, rands = logit.make_choices(
        choosers[purpose_cols], trace_label=trace_label, trace_choosers=choosers
    )

    if have_trace_targets:
        tracing.trace_df(
            choices, "%s.choices" % trace_label, columns=[None, "trip_purpose"]
        )
        tracing.trace_df(rands, "%s.rands" % trace_label, columns=[None, "rand"])

    choices = choices.map(pd.Series(purpose_cols))
    return choices


def run_trip_purpose(trips_df, estimator, chunk_size, trace_hh_id, trace_label):
    """
    trip purpose - main functionality separated from model step so it can be called iteratively

    For each intermediate stop on a tour (i.e. trip other than the last trip outbound or inbound)
    each trip is assigned a purpose based on an observed frequency distribution

    The distribution should always be segmented by tour purpose and tour direction. By default it is also
    segmented by person type. The join columns can be overwritten using the "probs_join_cols" parameter in
    the model settings. The model will attempt to segment by trip depart time as well if necessary
    and depart time ranges are specified in the probability lookup table.

    Returns
    -------
    purpose: pandas.Series of purpose (str) indexed by trip_id
    """

    # uniform across trip_purpose
    chunk_tag = "trip_purpose"

    model_settings_file_name = "trip_purpose.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    probs_join_cols = model_settings.get("probs_join_cols", PROBS_JOIN_COLUMNS)

    spec_file_name = model_settings.get("PROBS_SPEC", "trip_purpose_probs.csv")
    probs_spec = pd.read_csv(config.config_file_path(spec_file_name), comment="#")
    # FIXME for now, not really doing estimation for probabilistic model - just overwriting choices
    # besides, it isn't clear that named coefficients would be helpful if we had some form of estimation
    # coefficients_df = simulate.read_model_coefficients(model_settings)
    # probs_spec = map_coefficients(probs_spec, coefficients_df)

    if estimator:
        estimator.write_spec(model_settings, tag="PROBS_SPEC")
        estimator.write_model_settings(model_settings, model_settings_file_name)
        # estimator.write_coefficients(coefficients_df, model_settings)

    result_list = []

    # - last trip of outbound tour gets primary_purpose
    last_trip = trips_df.trip_num == trips_df.trip_count
    purpose = trips_df.primary_purpose[last_trip & trips_df.outbound]
    result_list.append(purpose)
    logger.info("assign purpose to %s last outbound trips", purpose.shape[0])

    # - last trip of inbound tour gets home (or work for atwork subtours)
    purpose = trips_df.primary_purpose[last_trip & ~trips_df.outbound]
    purpose = pd.Series(
        np.where(purpose == "atwork", "work", "home"), index=purpose.index
    )
    result_list.append(purpose)
    logger.info("assign purpose to %s last inbound trips", purpose.shape[0])

    # - intermediate stops (non-last trips) purpose assigned by probability table
    trips_df = trips_df[~last_trip]
    logger.info("assign purpose to %s intermediate trips", trips_df.shape[0])

    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:
        locals_dict = config.get_model_constants(model_settings)
        expressions.assign_columns(
            df=trips_df,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label,
        )

    use_depart_time = model_settings.get("use_depart_time", True)

    for i, trips_chunk, chunk_trace_label in chunk.adaptive_chunked_choosers(
        trips_df, chunk_size, chunk_tag, trace_label
    ):
        choices = choose_intermediate_trip_purpose(
            trips_chunk,
            probs_spec,
            estimator,
            probs_join_cols=probs_join_cols,
            use_depart_time=use_depart_time,
            trace_hh_id=trace_hh_id,
            trace_label=chunk_trace_label,
        )

        result_list.append(choices)

        chunk.log_df(trace_label, f"result_list", result_list)

    if len(result_list) > 1:
        choices = pd.concat(result_list)

    return choices


@inject.step()
def trip_purpose(trips, chunk_size, trace_hh_id):

    """
    trip purpose model step - calls run_trip_purpose to run the actual model

    adds purpose column to trips
    """
    trace_label = "trip_purpose"

    trips_df = trips.to_frame()

    estimator = estimation.manager.begin_estimation("trip_purpose")
    if estimator:
        chooser_cols_for_estimation = [
            "person_id",
            "household_id",
            "tour_id",
            "trip_num",
        ]
        estimator.write_choosers(trips_df[chooser_cols_for_estimation])

    choices = run_trip_purpose(
        trips_df,
        estimator,
        chunk_size=chunk_size,
        trace_hh_id=trace_hh_id,
        trace_label=trace_label,
    )

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(
            choices, "trips", "purpose"
        )  # override choices
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    trips_df["purpose"] = choices

    # we should have assigned a purpose to all trips
    assert not trips_df.purpose.isnull().any()

    pipeline.replace_table("trips", trips_df)

    if trace_hh_id:
        tracing.trace_df(
            trips_df,
            label=trace_label,
            slicer="trip_id",
            index_label="trip_id",
            warn_if_empty=True,
        )



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import (
    config,
    expressions,
    inject,
    logit,
    pipeline,
    simulate,
    tracing,
)
from activitysim.core.interaction_sample_simulate import interaction_sample_simulate
from activitysim.core.tracing import print_elapsed_time
from activitysim.core.util import assign_in_place

from .util import estimation

logger = logging.getLogger(__name__)

NO_DESTINATION = -1


def wrap_skims(model_settings):
    """
    wrap skims of trip destination using origin, dest column names from model settings.
    Various of these are used by destination_sample, compute_logsums, and destination_simulate
    so we create them all here with canonical names.

    Note that compute_logsums aliases their names so it can use the same equations to compute
    logsums from origin to alt_dest, and from alt_dest to primarly destination

    odt_skims - SkimStackWrapper: trip origin, trip alt_dest, time_of_day
    dot_skims - SkimStackWrapper: trip alt_dest, trip origin, time_of_day
    dpt_skims - SkimStackWrapper: trip alt_dest, trip primary_dest, time_of_day
    pdt_skims - SkimStackWrapper: trip primary_dest,trip alt_dest, time_of_day
    od_skims - SkimDictWrapper: trip origin, trip alt_dest
    dp_skims - SkimDictWrapper: trip alt_dest, trip primary_dest

    Parameters
    ----------
    model_settings

    Returns
    -------
        dict containing skims, keyed by canonical names relative to tour orientation
    """

    network_los = inject.get_injectable("network_los")
    skim_dict = network_los.get_default_skim_dict()

    origin = model_settings["TRIP_ORIGIN"]
    park_zone = model_settings["ALT_DEST_COL_NAME"]
    destination = model_settings["TRIP_DESTINATION"]
    time_period = model_settings["TRIP_DEPARTURE_PERIOD"]

    skims = {
        "odt_skims": skim_dict.wrap_3d(
            orig_key=origin, dest_key=destination, dim3_key=time_period
        ),
        "dot_skims": skim_dict.wrap_3d(
            orig_key=destination, dest_key=origin, dim3_key=time_period
        ),
        "opt_skims": skim_dict.wrap_3d(
            orig_key=origin, dest_key=park_zone, dim3_key=time_period
        ),
        "pdt_skims": skim_dict.wrap_3d(
            orig_key=park_zone, dest_key=destination, dim3_key=time_period
        ),
        "od_skims": skim_dict.wrap(origin, destination),
        "do_skims": skim_dict.wrap(destination, origin),
        "op_skims": skim_dict.wrap(origin, park_zone),
        "pd_skims": skim_dict.wrap(park_zone, destination),
    }

    return skims


def get_spec_for_segment(model_settings, spec_name, segment):

    omnibus_spec = simulate.read_model_spec(file_name=model_settings[spec_name])

    spec = omnibus_spec[[segment]]

    # might as well ignore any spec rows with 0 utility
    spec = spec[spec.iloc[:, 0] != 0]
    assert spec.shape[0] > 0

    return spec


def parking_destination_simulate(
    segment_name,
    trips,
    destination_sample,
    model_settings,
    skims,
    chunk_size,
    trace_hh_id,
    trace_label,
):
    """
    Chose destination from destination_sample (with od_logsum and dp_logsum columns added)


    Returns
    -------
    choices - pandas.Series
        destination alt chosen
    """
    trace_label = tracing.extend_trace_label(trace_label, "trip_destination_simulate")

    spec = get_spec_for_segment(model_settings, "SPECIFICATION", segment_name)

    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]

    logger.info("Running trip_destination_simulate with %d trips", len(trips))

    locals_dict = config.get_model_constants(model_settings).copy()
    locals_dict.update(skims)

    parking_locations = interaction_sample_simulate(
        choosers=trips,
        alternatives=destination_sample,
        spec=spec,
        choice_column=alt_dest_col_name,
        want_logsums=False,
        allow_zero_probs=True,
        zero_prob_choice_val=NO_DESTINATION,
        skims=skims,
        locals_d=locals_dict,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name="parking_loc",
    )

    # drop any failed zero_prob destinations
    if (parking_locations == NO_DESTINATION).any():
        logger.debug(
            "dropping %s failed parking locations",
            (parking_locations == NO_DESTINATION).sum(),
        )
        parking_locations = parking_locations[parking_locations != NO_DESTINATION]

    return parking_locations


def choose_parking_location(
    segment_name,
    trips,
    alternatives,
    model_settings,
    want_sample_table,
    skims,
    chunk_size,
    trace_hh_id,
    trace_label,
):

    logger.info("choose_parking_location %s with %d trips", trace_label, trips.shape[0])

    t0 = print_elapsed_time()

    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]
    destination_sample = logit.interaction_dataset(
        trips, alternatives, alt_index_id=alt_dest_col_name
    )
    destination_sample.index = np.repeat(trips.index.values, len(alternatives))
    destination_sample.index.name = trips.index.name
    destination_sample = destination_sample[[alt_dest_col_name]].copy()

    # # - trip_destination_simulate
    destinations = parking_destination_simulate(
        segment_name=segment_name,
        trips=trips,
        destination_sample=destination_sample,
        model_settings=model_settings,
        skims=skims,
        chunk_size=chunk_size,
        trace_hh_id=trace_hh_id,
        trace_label=trace_label,
    )

    if want_sample_table:
        # FIXME - sample_table
        destination_sample.set_index(
            model_settings["ALT_DEST_COL_NAME"], append=True, inplace=True
        )
    else:
        destination_sample = None

    t0 = print_elapsed_time("%s.parking_location_simulate" % trace_label, t0)

    return destinations, destination_sample


def run_parking_destination(
    model_settings,
    trips,
    land_use,
    chunk_size,
    trace_hh_id,
    trace_label,
    fail_some_trips_for_testing=False,
):

    chooser_filter_column = model_settings.get("CHOOSER_FILTER_COLUMN_NAME")
    chooser_segment_column = model_settings.get("CHOOSER_SEGMENT_COLUMN_NAME")

    parking_location_column_name = model_settings["ALT_DEST_COL_NAME"]
    sample_table_name = model_settings.get("DEST_CHOICE_SAMPLE_TABLE_NAME")
    want_sample_table = (
        config.setting("want_dest_choice_sample_tables")
        and sample_table_name is not None
    )

    choosers = trips[trips[chooser_filter_column]]
    choosers = choosers.sort_index()

    # Placeholder for trips without a parking choice
    trips[parking_location_column_name] = -1

    skims = wrap_skims(model_settings)

    alt_column_filter_name = model_settings.get("ALTERNATIVE_FILTER_COLUMN_NAME")
    alternatives = land_use[land_use[alt_column_filter_name]]

    # don't need size terms in alternatives, just TAZ index
    alternatives = alternatives.drop(alternatives.columns, axis=1)
    alternatives.index.name = parking_location_column_name

    choices_list = []
    sample_list = []
    for segment_name, chooser_segment in choosers.groupby(chooser_segment_column):
        if chooser_segment.shape[0] == 0:
            logger.info(
                "%s skipping segment %s: no choosers", trace_label, segment_name
            )
            continue

        choices, destination_sample = choose_parking_location(
            segment_name,
            chooser_segment,
            alternatives,
            model_settings,
            want_sample_table,
            skims,
            chunk_size,
            trace_hh_id,
            trace_label=tracing.extend_trace_label(trace_label, segment_name),
        )

        choices_list.append(choices)
        if want_sample_table:
            assert destination_sample is not None
            sample_list.append(destination_sample)

    if len(choices_list) > 0:
        parking_df = pd.concat(choices_list)

        if fail_some_trips_for_testing:
            parking_df = parking_df.drop(parking_df.index[0])

        assign_in_place(trips, parking_df.to_frame(parking_location_column_name))
        trips[parking_location_column_name] = trips[
            parking_location_column_name
        ].fillna(-1)
    else:
        trips[parking_location_column_name] = -1

    save_sample_df = pd.concat(sample_list) if len(sample_list) > 0 else None

    return trips[parking_location_column_name], save_sample_df


@inject.step()
def parking_location(
    trips, trips_merged, land_use, network_los, chunk_size, trace_hh_id
):
    """
    Given a set of trips, each trip needs to have a parking location if
    it is eligible for remote parking.
    """

    trace_label = "parking_location"
    model_settings = config.read_model_settings("parking_location_choice.yaml")
    alt_destination_col_name = model_settings["ALT_DEST_COL_NAME"]

    preprocessor_settings = model_settings.get("PREPROCESSOR", None)

    trips_df = trips.to_frame()
    trips_merged_df = trips_merged.to_frame()
    land_use_df = land_use.to_frame()

    locals_dict = {"network_los": network_los}
    locals_dict.update(config.get_model_constants(model_settings))

    if preprocessor_settings:
        expressions.assign_columns(
            df=trips_merged_df,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label,
        )

    parking_locations, save_sample_df = run_parking_destination(
        model_settings,
        trips_merged_df,
        land_use_df,
        chunk_size=chunk_size,
        trace_hh_id=trace_hh_id,
        trace_label=trace_label,
    )

    assign_in_place(trips_df, parking_locations.to_frame(alt_destination_col_name))

    pipeline.replace_table("trips", trips_df)

    if trace_hh_id:
        tracing.trace_df(
            trips_df,
            label=trace_label,
            slicer="trip_id",
            index_label="trip_id",
            warn_if_empty=True,
        )

    if save_sample_df is not None:
        assert len(save_sample_df.index.get_level_values(0).unique()) == len(
            trips_df[trips_df.trip_num < trips_df.trip_count]
        )

        sample_table_name = model_settings.get("PARKING_LOCATION_SAMPLE_TABLE_NAME")
        assert sample_table_name is not None

        logger.info(
            "adding %s samples to %s" % (len(save_sample_df), sample_table_name)
        )

        # lest they try to put tour samples into the same table
        if pipeline.is_table(sample_table_name):
            raise RuntimeError("sample table %s already exists" % sample_table_name)
        pipeline.extend_table(sample_table_name, save_sample_df)



import pandas as pd
import yaml


def read_results(json_file):
    """
    Read benchmarking results from a single commit on a single machine.

    Parameters
    ----------
    json_file : str
        Path to the json file containing the target results.

    Returns
    -------
    pandas.DataFrame
    """
    out_data = {}
    with open(json_file, "rt") as f:
        in_data = yaml.safe_load(f)
    for k, v in in_data["results"].items():
        if v is None:
            continue
        m, c, _ = k.split(".")
        if m not in out_data:
            out_data[m] = {}
        out_data[m][c] = v["result"][0]
    return pd.DataFrame(out_data)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import assign, config, inject, simulate, tracing
from activitysim.core.util import assign_in_place

logger = logging.getLogger(__name__)


def compute_columns(df, model_settings, locals_dict={}, trace_label=None):
    """
    Evaluate expressions_spec in context of df, with optional additional pipeline tables in locals

    Parameters
    ----------
    df : pandas DataFrame
        or if None, expect name of pipeline table to be specified by DF in model_settings
    model_settings : dict or str
        dict with keys:
            DF - df_alias and (additionally, if df is None) name of pipeline table to load as df
            SPEC - name of expressions file (csv suffix optional) if different from model_settings
            TABLES - list of pipeline tables to load and make available as (read only) locals
        str:
            name of yaml file in configs_dir to load dict from
    locals_dict : dict
        dict of locals (e.g. utility functions) to add to the execution environment
    trace_label

    Returns
    -------
    results: pandas.DataFrame
        one column for each expression (except temps with ALL_CAP target names)
        same index as df
    """

    if isinstance(model_settings, str):
        model_settings_name = model_settings
        model_settings = config.read_model_settings("%s.yaml" % model_settings)
        assert model_settings, "Found no model settings for %s" % model_settings_name
    else:
        model_settings_name = "dict"
        assert isinstance(model_settings, dict)

    assert "DF" in model_settings, "Expected to find 'DF' in %s" % model_settings_name

    df_name = model_settings.get("DF")
    helper_table_names = model_settings.get("TABLES", [])
    expressions_spec_name = model_settings.get("SPEC", None)

    assert expressions_spec_name is not None, (
        "Expected to find 'SPEC' in %s" % model_settings_name
    )

    trace_label = tracing.extend_trace_label(trace_label or "", expressions_spec_name)

    if not expressions_spec_name.endswith(".csv"):
        expressions_spec_name = "%s.csv" % expressions_spec_name
    logger.debug(
        f"{trace_label} compute_columns using expression spec file {expressions_spec_name}"
    )
    expressions_spec = assign.read_assignment_spec(
        config.config_file_path(expressions_spec_name)
    )

    assert expressions_spec.shape[0] > 0, (
        "Expected to find some assignment expressions in %s" % expressions_spec_name
    )

    tables = {t: inject.get_table(t).to_frame() for t in helper_table_names}

    # if df was passed in, df might be a slice, or any other table, but DF is it's local alias
    assert df_name not in tables, "Did not expect to find df '%s' in TABLES" % df_name
    tables[df_name] = df

    # be nice and also give it to them as df?
    tables["df"] = df

    _locals_dict = assign.local_utilities()
    _locals_dict.update(locals_dict)
    _locals_dict.update(tables)

    # FIXME a number of asim model preprocessors want skim_dict - should they request it in model_settings.TABLES?
    _locals_dict.update(
        {
            # 'los': inject.get_injectable('network_los', None),
            "skim_dict": inject.get_injectable("skim_dict", None),
        }
    )

    results, trace_results, trace_assigned_locals = assign.assign_variables(
        expressions_spec, df, _locals_dict, trace_rows=tracing.trace_targets(df)
    )

    if trace_results is not None:
        tracing.trace_df(trace_results, label=trace_label, slicer="NONE")

    if trace_assigned_locals:
        tracing.write_csv(trace_assigned_locals, file_name="%s_locals" % trace_label)

    return results


def assign_columns(df, model_settings, locals_dict={}, trace_label=None):
    """
    Evaluate expressions in context of df and assign resulting target columns to df

    Can add new or modify existing columns (if target same as existing df column name)

    Parameters - same as for compute_columns except df must not be None
    Returns - nothing since we modify df in place
    """

    assert df is not None
    assert model_settings is not None

    results = compute_columns(df, model_settings, locals_dict, trace_label)

    assign_in_place(df, results)


# ##################################################################################################
# helpers
# ##################################################################################################


def annotate_preprocessors(df, locals_dict, skims, model_settings, trace_label):

    locals_d = {}
    locals_d.update(locals_dict)
    locals_d.update(skims)

    preprocessor_settings = model_settings.get("preprocessor", [])
    if not isinstance(preprocessor_settings, list):
        assert isinstance(preprocessor_settings, dict)
        preprocessor_settings = [preprocessor_settings]

    simulate.set_skim_wrapper_targets(df, skims)

    for model_settings in preprocessor_settings:

        results = compute_columns(
            df=df,
            model_settings=model_settings,
            locals_dict=locals_d,
            trace_label=trace_label,
        )

        assign_in_place(df, results)


def filter_chooser_columns(choosers, chooser_columns):

    missing_columns = [c for c in chooser_columns if c not in choosers]
    if missing_columns:
        logger.debug("filter_chooser_columns missing_columns %s" % missing_columns)

    # ignore any columns not appearing in choosers df
    chooser_columns = [c for c in chooser_columns if c in choosers]

    choosers = choosers[chooser_columns]
    return choosers



# ActivitySim
# See full license in LICENSE.txt.
import logging
import logging.config
import os.path

import numpy as np
import pandas as pd
import pytest

from .. import assign, config, inject, tracing


def setup_function():
    configs_dir = os.path.join(os.path.dirname(__file__), "configs")
    inject.add_injectable("configs_dir", configs_dir)


def close_handlers():

    loggers = logging.Logger.manager.loggerDict
    for name in loggers:
        logger = logging.getLogger(name)
        logger.handlers = []
        logger.propagate = True
        logger.setLevel(logging.NOTSET)


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


@pytest.fixture(scope="module")
def data_dir():
    return os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def spec_name(data_dir):
    return os.path.join(data_dir, "assignment_spec.csv")


@pytest.fixture(scope="module")
def data_name(data_dir):
    return os.path.join(data_dir, "data.csv")


@pytest.fixture(scope="module")
def data(data_name):
    return pd.read_csv(data_name)


def test_read_model_spec():
    spec = assign.read_assignment_spec(config.config_file_path("assignment_spec.csv"))

    assert len(spec) == 8

    assert list(spec.columns) == ["description", "target", "expression"]


def test_assign_variables(capsys, data):

    spec = assign.read_assignment_spec(config.config_file_path("assignment_spec.csv"))

    locals_d = {"CONSTANT": 7, "_shadow": 99}

    results, trace_results, trace_assigned_locals = assign.assign_variables(
        spec, data, locals_d, trace_rows=None
    )

    print(results)

    assert list(results.columns) == ["target1", "target2", "target3"]
    assert list(results.target1) == [True, False, False]
    assert list(results.target2) == [53, 53, 55]
    assert list(results.target3) == [530, 530, 550]
    assert trace_results is None
    assert trace_assigned_locals is None

    trace_rows = [False, True, False]

    results, trace_results, trace_assigned_locals = assign.assign_variables(
        spec, data, locals_d, trace_rows=trace_rows
    )

    # should get same results as before
    assert list(results.target3) == [530, 530, 550]

    # should assign trace_results for second row in data
    print(trace_results)

    assert trace_results is not None
    assert "_scalar" in trace_results.columns
    assert list(trace_results["_scalar"]) == [42]

    # shadow should have been assigned
    assert list(trace_results["_shadow"]) == [1]
    assert list(trace_results["_temp"]) == [9]
    assert list(trace_results["target3"]) == [530]

    print("trace_assigned_locals", trace_assigned_locals)
    assert trace_assigned_locals["_DF_COL_NAME"] == "thing2"

    # shouldn't have been changed even though it was a target
    assert locals_d["_shadow"] == 99

    out, err = capsys.readouterr()


def test_assign_variables_aliased(capsys, data):

    spec = assign.read_assignment_spec(
        config.config_file_path("assignment_spec_alias_df.csv")
    )

    locals_d = {"CONSTANT": 7, "_shadow": 99}

    trace_rows = [False, True, False]

    results, trace_results, trace_assigned_locals = assign.assign_variables(
        spec, data, locals_d, df_alias="aliased_df", trace_rows=trace_rows
    )

    print(results)

    assert list(results.columns) == ["target1", "target2", "target3"]
    assert list(results.target1) == [True, False, False]
    assert list(results.target2) == [53, 53, 55]
    assert list(results.target3) == [530, 530, 550]

    # should assign trace_results for second row in data
    print(trace_results)

    assert trace_results is not None
    assert "_scalar" in trace_results.columns
    assert list(trace_results["_scalar"]) == [42]

    # shadow should have been assigned
    assert list(trace_results["_shadow"]) == [1]
    assert list(trace_results["_temp"]) == [9]
    assert list(trace_results["target3"]) == [530]

    assert locals_d["_shadow"] == 99

    out, err = capsys.readouterr()


def test_assign_variables_failing(capsys, data):

    close_handlers()

    output_dir = os.path.join(os.path.dirname(__file__), "output")
    inject.add_injectable("output_dir", output_dir)

    tracing.config_logger(basic=True)

    spec = assign.read_assignment_spec(
        config.config_file_path("assignment_spec_failing.csv")
    )

    locals_d = {
        "CONSTANT": 7,
        "_shadow": 99,
        "log": np.log,
    }

    with pytest.raises(NameError) as excinfo:
        results, trace_results = assign.assign_variables(
            spec, data, locals_d, trace_rows=None
        )

    out, err = capsys.readouterr()
    # don't consume output
    print(out)

    # undefined variable should raise error
    assert "'undefined_variable' is not defined" in str(excinfo.value)



# ActivitySim
# See full license in LICENSE.txt.
import logging
import os.path

import pandas as pd
import pytest

from .. import inject, tracing


def close_handlers():

    loggers = logging.Logger.manager.loggerDict
    for name in loggers:
        logger = logging.getLogger(name)
        logger.handlers = []
        logger.propagate = True
        logger.setLevel(logging.NOTSET)


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


def add_canonical_dirs():

    inject.clear_cache()

    configs_dir = os.path.join(os.path.dirname(__file__), "configs")
    inject.add_injectable("configs_dir", configs_dir)

    output_dir = os.path.join(os.path.dirname(__file__), "output")
    inject.add_injectable("output_dir", output_dir)


def test_config_logger(capsys):

    add_canonical_dirs()

    tracing.config_logger()

    logger = logging.getLogger("activitysim")

    file_handlers = [h for h in logger.handlers if type(h) is logging.FileHandler]
    assert len(file_handlers) == 1
    asim_logger_baseFilename = file_handlers[0].baseFilename

    print("handlers:", logger.handlers)

    logger.info("test_config_logger")
    logger.info("log_info")
    logger.warning("log_warn1")

    out, err = capsys.readouterr()

    # don't consume output
    print(out)

    assert "could not find conf file" not in out
    assert "log_warn1" in out
    assert "log_info" not in out

    close_handlers()

    logger = logging.getLogger(__name__)
    logger.warning("log_warn2")

    with open(asim_logger_baseFilename, "r") as content_file:
        content = content_file.read()
        print(content)
    assert "log_warn1" in content
    assert "log_warn2" not in content


def test_print_summary(capsys):

    add_canonical_dirs()

    tracing.config_logger()

    tracing.print_summary(
        "label", df=pd.DataFrame(), describe=False, value_counts=False
    )

    out, err = capsys.readouterr()

    # don't consume output
    print(out)

    assert "print_summary neither value_counts nor describe" in out

    close_handlers()


def test_register_households(capsys):

    add_canonical_dirs()

    tracing.config_logger()

    df = pd.DataFrame({"zort": ["a", "b", "c"]}, index=[1, 2, 3])

    inject.add_injectable("traceable_tables", ["households"])
    inject.add_injectable("trace_hh_id", 5)

    tracing.register_traceable_table("households", df)
    out, err = capsys.readouterr()
    # print out   # don't consume output

    assert "Can't register table 'households' without index name" in out

    df.index.name = "household_id"
    tracing.register_traceable_table("households", df)
    out, err = capsys.readouterr()
    # print out   # don't consume output

    # should warn that household id not in index
    assert "trace_hh_id 5 not in dataframe" in out

    close_handlers()


def test_register_tours(capsys):

    add_canonical_dirs()

    tracing.config_logger()

    inject.add_injectable("traceable_tables", ["households", "tours"])

    # in case another test injected this
    inject.add_injectable("trace_tours", [])
    inject.add_injectable(
        "trace_hh_id", 3
    )  # need this or register_traceable_table is a nop

    tours_df = pd.DataFrame({"zort": ["a", "b", "c"]}, index=[10, 11, 12])
    tours_df.index.name = "tour_id"

    tracing.register_traceable_table("tours", tours_df)

    out, err = capsys.readouterr()
    assert (
        "can't find a registered table to slice table 'tours' index name 'tour_id'"
        in out
    )

    inject.add_injectable("trace_hh_id", 3)
    households_df = pd.DataFrame({"dzing": ["a", "b", "c"]}, index=[1, 2, 3])
    households_df.index.name = "household_id"
    tracing.register_traceable_table("households", households_df)

    tracing.register_traceable_table("tours", tours_df)

    out, err = capsys.readouterr()
    # print out  # don't consume output
    assert "can't find a registered table to slice table 'tours'" in out

    tours_df["household_id"] = [1, 5, 3]

    tracing.register_traceable_table("tours", tours_df)

    out, err = capsys.readouterr()
    print(out)  # don't consume output

    # should be tracing tour with tour_id 3
    traceable_table_ids = inject.get_injectable("traceable_table_ids")
    assert traceable_table_ids["tours"] == [12]

    close_handlers()


def test_write_csv(capsys):

    add_canonical_dirs()

    tracing.config_logger()

    # should complain if df not a DataFrame or Series
    tracing.write_csv(df="not a df or series", file_name="baddie")

    out, err = capsys.readouterr()

    print(out)  # don't consume output

    assert "unexpected type" in out

    close_handlers()


def test_slice_ids():

    df = pd.DataFrame({"household_id": [1, 2, 3]}, index=[11, 12, 13])

    # slice by named column
    sliced_df = tracing.slice_ids(df, [1, 3, 6], column="household_id")
    assert len(sliced_df.index) == 2

    # slice by index
    sliced_df = tracing.slice_ids(df, [6, 12], column=None)
    assert len(sliced_df.index) == 1

    # attempt to slice by non-existent column
    with pytest.raises(RuntimeError) as excinfo:
        sliced_df = tracing.slice_ids(df, [5, 6], column="baddie")
    assert "slice_ids slicer column 'baddie' not in dataframe" in str(excinfo.value)


def test_basic(capsys):

    close_handlers()

    configs_dir = os.path.join(os.path.dirname(__file__), "configs")
    inject.add_injectable("configs_dir", configs_dir)

    output_dir = os.path.join(os.path.dirname(__file__), "output")
    inject.add_injectable("output_dir", output_dir)

    # remove existing handlers or basicConfig is a NOP
    logging.getLogger().handlers = []

    tracing.config_logger(basic=True)

    logger = logging.getLogger()
    file_handlers = [h for h in logger.handlers if type(h) is logging.FileHandler]
    assert len(file_handlers) == 0

    logger = logging.getLogger("activitysim")

    logger.info("test_basic")
    logger.debug("log_debug")
    logger.info("log_info")
    logger.warning("log_warn")

    out, err = capsys.readouterr()

    # don't consume output
    print(out)

    assert "log_warn" in out
    assert "log_info" in out
    assert "log_debug" not in out

    close_handlers()



# ActivitySim
# See full license in LICENSE.txt.

import logging
import os
import warnings

import pandas as pd

from activitysim.core import config, inject, mem, util

logger = logging.getLogger(__name__)


def canonical_table_index_name(table_name):
    table_index_names = inject.get_injectable("canonical_table_index_names", None)
    return table_index_names and table_index_names.get(table_name, None)


def read_input_table(tablename, required=True):
    """Reads input table name and returns cleaned DataFrame.

    Uses settings found in input_table_list in global settings file

    Parameters
    ----------
    tablename : string

    Returns
    -------
    pandas DataFrame
    """
    table_list = config.setting("input_table_list")
    assert table_list is not None, "no input_table_list found in settings"

    table_info = None
    for info in table_list:
        if info["tablename"] == tablename:
            table_info = info

    if table_info is not None:
        df = read_from_table_info(table_info)
    else:
        if required:
            raise RuntimeError(
                f"could not find info for for tablename {tablename} in settings file"
            )
        df = None

    return df


def read_from_table_info(table_info):
    """
    Read input text files and return cleaned up DataFrame.

    table_info is a dictionary that specifies the following input params.

    See input_table_list in settings.yaml in the example folder for a working example

    +--------------+----------------------------------------------------------+
    | key          | description                                              |
    +==============+=========================================+================+
    | tablename    | name of pipeline table in which to store dataframe       |
    +--------------+----------------------------------------------------------+
    | filename     | name of csv file to read (in data_dir)                   |
    +--------------+----------------------------------------------------------+
    | column_map   | list of input columns to rename from_name: to_name       |
    +--------------+----------------------------------------------------------+
    | index_col    | name of column to set as dataframe index column          |
    +--------------+----------------------------------------------------------+
    | drop_columns | list of column names of columns to drop                  |
    +--------------+----------------------------------------------------------+
    | h5_tablename | name of target table in HDF5 file                        |
    +--------------+----------------------------------------------------------+

    """
    input_store = config.setting("input_store", None)
    create_input_store = config.setting("create_input_store", default=False)

    tablename = table_info.get("tablename")
    data_filename = table_info.get("filename", input_store)
    h5_tablename = table_info.get("h5_tablename") or tablename
    drop_columns = table_info.get("drop_columns", None)
    column_map = table_info.get("column_map", None)
    keep_columns = table_info.get("keep_columns", None)
    rename_columns = table_info.get("rename_columns", None)
    csv_dtypes = table_info.get("dtypes", {})

    # don't require a redundant index_col directive for canonical tables
    # but allow explicit disabling of assignment of index col for canonical tables, in which case, presumably,
    # the canonical index will be assigned in a subsequent initialization step (e.g. initialize_tours)
    canonical_index_col = canonical_table_index_name(tablename)

    # if there is an explicit index_col entry in table_info
    if "index_col" in table_info:
        # honor explicit index_col unless it conflicts with canonical name

        index_col = table_info["index_col"]

        if canonical_index_col:
            if index_col:
                # if there is a non-empty index_col directive, it should be for canonical_table_index_name
                assert (
                    index_col == canonical_index_col
                ), f"{tablename} index_col {table_info.get('index_col')} should be {index_col}"
            else:
                logger.info(
                    f"Not assigning canonical index_col {tablename}.{canonical_index_col} "
                    f"because settings file index_col directive is explicitly None."
                )

        #  if there is an index_col directive for a canonical table, it should be for canonical_table_index_name

    else:
        # otherwise default is to use canonical index name for known tables, and no index for unknown tables
        index_col = canonical_index_col

    assert tablename is not None, "no tablename provided"
    assert data_filename is not None, "no input file provided"

    data_file_path = config.data_file_path(data_filename)

    df = _read_input_file(
        data_file_path, h5_tablename=h5_tablename, csv_dtypes=csv_dtypes
    )

    # logger.debug('raw %s table columns: %s' % (tablename, df.columns.values))
    logger.debug("raw %s table size: %s" % (tablename, util.df_size(df)))

    if create_input_store:
        h5_filepath = config.output_file_path("input_data.h5")
        logger.info("writing %s to %s" % (h5_tablename, h5_filepath))
        df.to_hdf(h5_filepath, key=h5_tablename, mode="a")

        csv_dir = config.output_file_path("input_data")
        if not os.path.exists(csv_dir):
            os.makedirs(csv_dir)  # make directory if needed
        df.to_csv(os.path.join(csv_dir, "%s.csv" % tablename), index=False)

    if drop_columns:
        logger.debug("dropping columns: %s" % drop_columns)
        df.drop(columns=drop_columns, inplace=True, errors="ignore")

    if column_map:
        warnings.warn(
            "table_inf option 'column_map' renamed 'rename_columns'"
            "Support for 'column_map' will be removed in future versions.",
            FutureWarning,
        )
        logger.debug("renaming columns: %s" % column_map)
        df.rename(columns=column_map, inplace=True)

    # rename columns first, so keep_columns can be a stable list of expected/required columns
    if rename_columns:
        logger.debug("renaming columns: %s" % rename_columns)
        df.rename(columns=rename_columns, inplace=True)

    # set index
    if index_col is not None:
        if index_col in df.columns:
            assert not df.duplicated(index_col).any()
            if canonical_index_col:
                # we expect canonical indexes to be integer-valued
                assert (
                    df[index_col] == df[index_col].astype(int)
                ).all(), f"Index col '{index_col}' has non-integer values"
                df[index_col] = df[index_col].astype(int)
            df.set_index(index_col, inplace=True)
        else:
            # FIXME not sure we want to do this. More likely they omitted index col than that they want to name it?
            # df.index.names = [index_col]
            logger.error(
                f"index_col '{index_col}' specified in configs but not in {tablename} table!"
            )
            logger.error(f"{tablename} columns are: {list(df.columns)}")
            raise RuntimeError(f"index_col '{index_col}' not in {tablename} table!")

    if keep_columns:
        logger.debug("keeping columns: %s" % keep_columns)
        if not set(keep_columns).issubset(set(df.columns)):
            logger.error(
                f"Required columns missing from {tablename} table: "
                f"{list(set(keep_columns).difference(set(df.columns)))}"
            )
            logger.error(f"{tablename} table has columns: {list(df.columns)}")
            raise RuntimeError(f"Required columns missing from {tablename} table")

        df = df[keep_columns]

    if df.columns.duplicated().any():
        duplicate_column_names = (
            df.columns[df.columns.duplicated(keep=False)].unique().to_list()
        )
        assert (
            not df.columns.duplicated().any()
        ), f"duplicate columns names in {tablename}: {duplicate_column_names}"

    logger.debug("%s table columns: %s" % (tablename, df.columns.values))
    logger.debug("%s table size: %s" % (tablename, util.df_size(df)))
    logger.debug("%s index name: %s" % (tablename, df.index.name))

    return df


def _read_input_file(filepath, h5_tablename=None, csv_dtypes=None):
    assert os.path.exists(filepath), "input file not found: %s" % filepath

    if filepath.endswith(".csv") or filepath.endswith(".csv.gz"):
        return _read_csv_with_fallback_encoding(filepath, csv_dtypes)

    if filepath.endswith(".h5"):
        assert h5_tablename is not None, "must provide a tablename to read HDF5 table"
        logger.info("reading %s table from %s" % (h5_tablename, filepath))
        return pd.read_hdf(filepath, h5_tablename)

    raise IOError(
        "Unsupported file type: %s. "
        "ActivitySim supports CSV and HDF5 files only" % filepath
    )


def _read_csv_with_fallback_encoding(filepath, dtypes=None):
    """read a CSV to a pandas DataFrame using default utf-8 encoding,
    but try alternate Windows-compatible cp1252 if unicode fails

    """

    try:
        logger.info("Reading CSV file %s" % filepath)
        df = pd.read_csv(filepath, comment="#", dtype=dtypes)
    except UnicodeDecodeError:
        logger.warning(
            "Reading %s with default utf-8 encoding failed, trying cp1252 instead",
            filepath,
        )
        df = pd.read_csv(filepath, comment="#", encoding="cp1252", dtype=dtypes)

    if dtypes:
        # although the dtype argument suppresses the DtypeWarning, it does not coerce recognized types (e.g. int)
        for c, dtype in dtypes.items():
            df[c] = df[c].astype(dtype)

    return df
